import sqlite3
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.time import Time

from tdastro.astro_utils.mag_flux import mag2flux
from tdastro.astro_utils.noise_model import poisson_bandflux_std
from tdastro.consts import GAUSS_EFF_AREA2FWHM_SQ
from tdastro.opsim.opsim import OpSim

ZTFCAM_PIXEL_SCALE = 1.01
"""The pixel scale for the ZTF camera in arcseconds per pixel."""

_ztfcam_readout_noise = 8
"""The standard deviation of the count of readout electrons per pixel for the ZTF camera."""

_ztfcam_dark_current = 0.0
"""The dark current for the ZTF camera in electrons per second per pixel."""

_ztfcam_view_radius = 2.735
"""The angular radius of the observation field (in degrees). ZTF FOV is 47 deg^2"""

_ztfcam_ccd_gain = 6.2
"""CCD gain (in e-/ADU)"""


def calculate_ztf_zero_points(
    maglim=None,
    sky=None,
    fwhm=None,
    gain=_ztfcam_ccd_gain,
    readnoise=_ztfcam_readout_noise,
    darkcurrent=_ztfcam_dark_current,
    exptime=None,
    nexposure=1,
):
    """
    Calculate zero points based on the 5-sigma mag limit.

    snr = flux/fluxerr
    fluxerr = sqrt(flux + sky*npix*gain + readnoise**2*nexposure*npix + darkcurrent*npix*exptime*nexposure)
    5 = flux/fluxerr
    25 = flux**2/(flux + sky*npix*Gain + readnoise**2*nexposure*npix + darkcurrent*npix*exptime*nexposure)
    flux**2 - 25*flux -25*( sky*npix*Gain
                            + readnoise**2*nexposure*npix
                            + darkcurrent*npix*exptime*nexposure)
                        = 0
    flux = 12.5 + 0.5*sqrt(625
                            + 100( sky*npix*Gain
                            + readnoise**2*nexposure*npix
                            + darkcurrent*npix*exptime*nexposure) )
    zp = 2.5log(flux) + maglim

    Parameters
    ----------
    maglim : float or ndarray
        Five-sigma magnitude limit.
    sky : float or ndarry
        Sky background in ADU/pixel.
    fwhm : float or ndarray
        PSF in pixels.
    gain : float or ndarray; default is _ztfcam_ccd_gain
        CCD gain.
    readnoise : float or ndarray; default is _ztfcam_readout_noise
        Read noise (in e-/pixel).
    darkcurrent : float or ndarray; default is _ztfcam_dark_current
        Dark current (in e-/pixel/second).
    exptime : float or ndarray
        Exposure time (in seconds).
    nexposure : int or ndarray
        Number of exposure.

    Returns
    -------
    zp: float or ndarray
        Instrument zero point (that converts 1 e- to magnitude).
    """

    npix = 2.266 * fwhm**2  # =4*pi*sigma**2=pi/2/ln2 * FWHM**2
    flux_at_5sigma_limit = 12.5 + 2.5 * np.sqrt(
        25.0
        + 4.0
        * (sky * npix * gain + readnoise**2 * nexposure * npix + darkcurrent * npix * exptime * nexposure)
    )
    zp = 2.5 * np.log10(flux_at_5sigma_limit) + maglim

    return zp


class ZTFOpsim(OpSim):
    """A subclass for ZTF exposure table.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the OpSim information.
    colmap : dict
        A mapping of short column names to their names in the underlying table.
        Defaults to the Rubin OpSim column names, stored in _default_colnames.
    **kwargs : dict
        Additional keyword arguments to pass to the OpSim constructor. This includes overrides
        for survey parameters such as:
        - dark_current : The dark current for the camera in electrons per second per pixel.
        - gain: The CCD gain (in e-/ADU).
        - pixel_scale: The pixel scale for the camera in arcseconds per pixel.
        - radius: The angular radius of the observations (in degrees).
        - read_noise: The standard deviation of the count of readout electrons per pixel.

    Attributes
    ----------
    table : pandas.core.frame.DataFrame
        The table with all the OpSim information.
    colmap : dict
        A mapping of short column names to their names in the underlying table.
    _kd_tree : scipy.spatial.KDTree or None
        A kd_tree of the OpSim pointings for fast spatial queries. We use the scipy
        kd-tree instead of astropy's functions so we can directly control caching.
    pixel_scale : float or None, optional
        The pixel scale for the ZTF camera in arcseconds per pixel.
    read_noise : float or None, optional
        The readout noise for the ZTF camera in electrons per pixel.
    dark_current : float or None, optional
        The dark current for the ZTF camera in electrons per second per pixel.
    radius : float
        The angular radius of the observations (in degrees).
    """

    # Default column names for the ZTF survey data.
    _default_colnames = {
        "maglim": "maglim",
        "sky": "scibckgnd",
        "fwhm": "fwhm",
        "dec": "dec",
        "exptime": "exptime",
        "filter": "filter",
        "ra": "ra",
        "time": "obsmjd",
        "zp": "zp_nJy",  # We add this column to the table
    }

    # Default survey values.
    _default_survey_values = {
        "dark_current": _ztfcam_dark_current,
        "gain": _ztfcam_ccd_gain,
        "pixel_scale": ZTFCAM_PIXEL_SCALE,
        "radius": _ztfcam_view_radius,
        "read_noise": _ztfcam_readout_noise,
    }

    def __init__(self, table, colmap=None, **kwargs):
        super().__init__(table, colmap=colmap, **kwargs)

        # Convert obsdate to mjd and add column
        obsdate = self.table[self.colmap.get("obsdate", "obsdate")].tolist()
        t = Time(obsdate, format="iso", scale="utc")
        mjd = t.mjd
        self.add_column(self.colmap.get("obsmjd", "obsmjd"), mjd, overwrite=True)

    def _assign_zero_points(self):
        """Assign instrumental zero points in ADU to the OpSim tables."""
        if not self.has_columns(["maglim", "sky", "fwhm", "exptime"]):
            raise ValueError(
                "OpSim does not include the columns needed to derive zero point "
                "information. Required columns: maglim, sky, fwhm and exptime."
            )

        # replace invalid values in table
        self.table = self.table.replace("", np.nan)
        self.table = self.table.dropna(subset=["fwhm"])

        zp_values = calculate_ztf_zero_points(
            maglim=self.table[self.colmap.get("maglim", "maglim")],
            sky=self.table[self.colmap.get("sky", "sky")],
            fwhm=self.table[self.colmap.get("fwhm", "fwhm")],
            exptime=self.table[self.colmap.get("exptime", "exptime")],
        )
        zp_nJy = mag2flux(zp_values)
        self.add_column(self.colmap.get("zp", "zp_nJy"), zp_nJy, overwrite=True)

    @classmethod
    def from_db(cls, filename, sql_query="SELECT * from exposures", colmap=None):
        """Create an OpSim object from the data in an opsim db file.

        Parameters
        ----------
        filename : str
            The name of the opsim db file.
        sql_query : str
            The SQL query to use when loading the table.
            Default = "SELECT * FROM observations"
        colmap : dict, optional
            A mapping of short column names to their names in the underlying table.
            If None then defaults to the ZTF column names.

        Returns
        -------
        opsim : OpSim
            A table with all of the pointing data.

        Raise
        -----
        FileNotFoundError if the file does not exist.
        ValueError if unable to load the table.
        """
        if colmap is None:
            colmap = cls._default_colnames

        if not Path(filename).is_file():
            raise FileNotFoundError(f"opsim file {filename} not found.")
        con = sqlite3.connect(f"file:{filename}?mode=ro", uri=True)

        # Read the table.
        try:
            opsim = pd.read_sql_query(sql_query, con)
        except Exception:
            raise ValueError("Opsim database read failed.") from None

        # Close the connection.
        con.close()

        return ZTFOpsim(opsim, colmap=colmap)

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the OpSim table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        observations = self.table.iloc[index]

        # By the effective FWHM definition, see
        # https://smtn-002.lsst.io/v/OPSIM-1171/index.html
        footprint = GAUSS_EFF_AREA2FWHM_SQ * observations["fwhm"] ** 2  # in pixels

        return poisson_bandflux_std(
            bandflux,  # nJy
            total_exposure_time=observations["exptime"],
            exposure_count=1,
            footprint=footprint,
            sky=observations["scibckgnd"] * self.gain,  # e-/pixel^2
            zp=observations["zp_nJy"],  # nJy
            readout_noise=self.read_noise,  # e-/pixel
            dark_current=self.dark_current,  # e-/second/pixel
        )


def create_random_ztf_opsim(num_obs, seed=None):
    """Create a random OpSim pointings drawn uniformly from (RA, dec).

    Parameters
    ----------
    num_obs : int
        The size of the OpSim to generate.
    seed : int
        The seed to used for random number generation. If None then
        uses a default random number generator.
        Default: None

    Returns
    -------
    opsim_data : OpSim
        The OpSim data structure.
    seed : int, optional
        The seed for the random number generator.
    """
    if num_obs <= 0:
        raise ValueError("Number of observations must be greater than zero.")

    rng = np.random.default_rng() if seed is None else np.random.default_rng(seed=seed)

    # Generate the (RA, dec) pairs uniformly on the surface of a sphere.
    ra = np.degrees(rng.uniform(0.0, 2.0 * np.pi, size=num_obs))
    dec = np.degrees(np.arccos(2.0 * rng.uniform(0.0, 1.0, size=num_obs) - 1.0) - (np.pi / 2.0))

    # Generate the information needed to compute zeropoint.
    maglim = rng.uniform(19.0, 21.0, size=num_obs)
    sky = rng.uniform(100.0, 200.0, size=num_obs)
    fwhm = rng.uniform(1.0, 3.0, size=num_obs)
    filter = rng.choice(["g", "r", "i"], size=num_obs)

    input_data = {
        "obsdate": ["2018-03-25 06:04:40.000"] * num_obs,
        "ra": ra,
        "dec": dec,
        "maglim": maglim,
        "scibckgnd": sky,
        "fwhm": fwhm,
        "filter": filter,
        "exptime": 30.0 * np.ones(num_obs),
    }

    opsim = ZTFOpsim(input_data)

    return opsim



"""A wrapper for querying dust maps and returning E(B-V) values.

This module supports a variety of external dustmaps libaries, but
was primarily designed to work with the dustmaps package:
https://github.com/gregreen/dustmaps
"""

import importlib
import logging
from pathlib import Path

import numpy as np
import pooch
from astropy.coordinates import SkyCoord
from citation_compass import CiteClass

from tdastro import _TDASTRO_BASE_DATA_DIR
from tdastro.base_models import FunctionNode


class DustEBV(FunctionNode):
    """A wrapper that queries a dust map and returns the ebv for each location.

    This wrapper is designed to work with multiple dust map implementtions
    by providing a stadard interface.

    Attributes
    ----------
    query_fn : function
        The function to query the dust map ebv value given (RA, dec).

    Parameters
    ----------
    query_fn : function
        The function to query the dust map ebv value given (RA, dec).
    RA : parameter
        The object's right ascension (in degrees).
    dec : parameter
        The object's declination (in degrees).
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    def __init__(self, query_fn, ra=None, dec=None, **kwargs):
        # RA and dec are required, but must be passed through as keyword
        # arguments to be used in the sampling function (hence the None default).
        if ra is None or dec is None:
            raise ValueError("RA and dec must be provided to query the dust map.")
        super().__init__(query_fn, ra=ra, dec=dec, outputs=["ebv"], **kwargs)


class DustmapWrapper(DustEBV, CiteClass):
    """A convenience wrapper for the dustmap package
    https://github.com/gregreen/dustmaps

    The DustmapWrapper is designed to take dust map objects from the dustmaps package,
    but can be used with any object that has a query(coords: SkyCoord) method.

    Parameters
    ----------
    dust_map : DustMap object
        The dust map. Since different dustmap's query function may produce different
        outputs, you should include the corresponding ebv_func to transform the result
        into ebv if needed.
    ebv_func : function, optional
        A function to translate the result of the dustmap query into an ebv.
    RA : parameter
        The object's right ascension (in degrees).
    dec : parameter
        The object's declination (in degrees).
    **kwargs : `dict`, optional
        Any additional keyword arguments.

    References
    ----------
    Green 2018, JOSS, 3(26), 695.
    https://github.com/gregreen/dustmaps
    """

    def __init__(self, dust_map, ra=None, dec=None, ebv_func=None, **kwargs):
        # RA and dec setters are passed through to the DustEBV init method.
        super().__init__(self.compute_ebv, ra=ra, dec=dec, **kwargs)
        self._ebv_func = ebv_func
        self._dust_map = dust_map

    def compute_ebv(self, ra, dec):
        """Compute the E(B-V) value for a given location.

        Parameters
        ----------
        ra : float or np.array
            The object's right ascension (in degrees).
        dec : float or np.array
            The object's declination (in degrees).

        Returns
        -------
        ebv : float or np.array
            The E(B-V) value or array of values.
        """
        if self._dust_map is None:
            raise ValueError("A dust map must be provided to compute E(B-V) values.")

        # Wrap the RA and dec in a SkyCoord object and call the dust map's query method.
        coord = SkyCoord(ra, dec, frame="icrs", unit="deg")
        dustmap_value = self._dust_map.query(coord)

        if self._ebv_func is not None:
            return self._ebv_func(dustmap_value)
        return dustmap_value


class ConstantHemisphereDustMap:
    """A DustMap with a constant value in each hemisphere for testing
    and debugging purposes.

    This class is designed to look like a DustMap object from the dustmaps package,
    so that it can be used to test DustmapWrapper.

    Attributes
    ----------
    north_ebv : `float`
        The DustMap's ebv value at all points in the Northern Hemisphere.
    south_ebv : `float`
        The DustMap's ebv value at all points in the Southern Hemisphere.
    """

    def __init__(self, north_ebv, south_ebv):
        self.north_ebv = north_ebv
        self.south_ebv = south_ebv

    def compute_ebv(self, ra, dec):
        """Compute the E(B-V) value for a given location.

        Parameters
        ----------
        ra : float or np.array
            The object's right ascension (in degrees).
        dec : float or np.array
            The object's declination (in degrees).

        Returns
        -------
        ebv : float or np.array
            The E(B-V) value or array of values.
        """
        if np.isscalar(ra):
            return self.north_ebv if dec >= 0 else self.south_ebv

        ebv = np.where(np.asarray(dec) < 0, self.south_ebv, self.north_ebv)
        return ebv

    def query(self, coords):
        """A query function to match the DustMap interface so that we can
        pass ConstantHemisphereDustMap into DustmapWrapper for testing purposes.

        Note
        ----
        This shouldn't be used in production code. Use the compute_ebv()
        function directly.

        Parameters
        ----------
        coords : SkyCoord
            The object's coordinates.

        Returns
        -------
        ebv : float or np.array
            The E(B-V) value or array of values.
        """
        return self.compute_ebv(coords.ra.deg, coords.dec.deg)


class SFDMap(DustEBV):
    """A dustmap using the sfdmap2 package.

    This does not need to be used with DustmapWrapper, but rather is
    a standalone ParameterizedNode.

    Note
    ----
    If the dustmap data is not present in "data/dustmaps/sfdmap2",
    it will automatically be downloaded there when the class is
    instantiated.

    Citations
    ---------
    Software https://github.com/AmpelAstro/sfdmap2
    Forked from https://github.com/kbarbary/sfdmap
    Dust map data from Schlegel, Finkbeiner and Davis (1998).

    Attributes
    ----------
    dustmap : sfdmap.SFDMap
        The dust map object.

    Parameters
    ----------
    data_dir : `str`, optional
        The directory containing the dust map data files.
        If None, the default directory will be used.
    RA : parameter
        The object's right ascension (in degrees).
    dec : parameter
        The object's declination (in degrees).
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    _default_map_dir = _TDASTRO_BASE_DATA_DIR / "dustmaps" / "sfdmap2"

    def __init__(self, data_dir=None, ra=None, dec=None, **kwargs):
        # Pass RA and dec through to DustEBV's init method.
        super().__init__(self.compute_ebv, ra=ra, dec=dec, **kwargs)
        self.dustmap = self._load_data(data_dir)

    def _data_files_exist(self, data_dir):
        """Check that the necessary data files exist in the given directory."""
        if not data_dir.exists() or not data_dir.is_dir():
            return False
        elif (data_dir / "SFD_dust_4096_ngp.fits").exists() is False:
            return False
        elif (data_dir / "SFD_dust_4096_sgp.fits").exists() is False:
            return False
        return True

    def _load_data(self, data_dir=None):
        """Load the dust map data files.

        Parameters
        ----------
        data_dir : `str`, optional
            The directory containing the dust map data files.
            If None, the default directory will be used.

        Returns
        -------
        dustmap : sfdmap.SFDMap
            The dust map object.
        """
        logger = logging.getLogger(__name__)

        # Check that we have the sfdmap2 package installed (since it is not installed by default).
        if importlib.util.find_spec("sfdmap2") is None:
            raise ImportError(
                "The sfdmap2 package is required to use the SFDMap effect and not installed by default. "
                "You can install sfdmap2 using `pip install sfdmap2`."
            )
        else:
            from sfdmap2 import sfdmap

        data_dir = Path(data_dir) if data_dir is not None else self._default_map_dir
        logger.debug(f"Loading SFD dust map data from {data_dir}")

        if not self._data_files_exist(data_dir):
            data_url = ("https://github.com/kbarbary/sfddata/archive/master.tar.gz",)
            logger.info(
                "SFD dust map data files not found.\n"
                f"Attempting to download from: {data_url}\n"
                f"to the directory {data_dir}"
            )

            # Create the data directory if it doesn't exist.
            if not data_dir.exists():
                data_dir.mkdir(parents=True)

            # Use pooch to download the data files and extract them to the data directory.
            pooch.retrieve(
                url="https://github.com/kbarbary/sfddata/archive/master.tar.gz",
                known_hash="95e8645dcdcbd4ad48398d44307741f550bdde95e8f438b2a3be0021723a4d7e",
                processor=pooch.Untar(extract_dir=data_dir),
            )
            data_dir = data_dir / "sfddata-master"

        # Check that the files were downloaded and extracted.
        if not self._data_files_exist(data_dir):
            raise ValueError(f"The SFD dust map data files are missing from {data_dir}.")

        return sfdmap.SFDMap(data_dir)

    def compute_ebv(self, ra, dec):
        """Compute the E(B-V) value for a given location.

        Parameters
        ----------
        ra : float or np.array
            The object's right ascension (in degrees).
        dec : float or np.array
            The object's declination (in degrees).

        Returns
        -------
        ebv : float or np.array
            The E(B-V) value or array of values.
        """
        ebv = self.dustmap.ebv(ra, dec)
        return ebv



"""The base PhysicalModel used for all sources."""

from os import urandom

import numpy as np

from tdastro.astro_utils.passbands import Passband
from tdastro.astro_utils.redshift import RedshiftDistFunc, obs_to_rest_times_waves, rest_to_obs_flux
from tdastro.base_models import ParameterizedNode
from tdastro.graph_state import GraphState


class PhysicalModel(ParameterizedNode):
    """A physical model of a source of flux.

    Physical models can have fixed attributes (where you need to create a new model or use
    a setter function to change them) and settable model parameters that can be passed functions
    or constants and are stored in the graph's (external) graph_state dictionary.

    Physical models also support adding and applying a variety of effects, such as redshift.

    Parameterized values include:
      * dec - The object's declination in degrees.
      * distance - The object's luminosity distance in pc.
      * ra - The object's right ascension in degrees.
      * redshift - The object's redshift.
      * t0 - The t0 of the zero phase (if applicable), date.

    Attributes
    ----------
    apply_redshift : bool
        Indicates whether to apply the redshift.
    rest_frame_effects : list of EffectModel
        A list of effects to apply in the rest frame.
    obs_frame_effects : list of EffectModel
        A list of effects to apply in the observer frame.
    wave_extrapolation : WaveExtrapolationModel, optional
        The extrapolation model to use for wavelengths that fall outside
        the model's defined bounds.  If None then the model will use all zeros.

    Parameters
    ----------
    ra : float
        The object's right ascension (in degrees)
    dec : float
        The object's declination (in degrees)
    redshift : float
        The object's redshift.
    t0 : float
        The phase offset in MJD. For non-time-varying phenomena, this has no effect.
    distance : float
        The object's luminosity distance (in pc). If no value is provided and
        a cosmology parameter is given, the model will try to derive from
        the redshift and the cosmology.
    wave_extrapolation : WaveExtrapolationModel, optional
        The extrapolation model to use for wavelengths that fall outside
        the model's defined bounds.  If None then the model will use all zeros.
    seed : int, optional
        The seed for a random number generator.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        ra=None,
        dec=None,
        redshift=None,
        t0=None,
        distance=None,
        wave_extrapolation=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Set the parameters for the model.
        self.add_parameter("ra", ra, allow_gradient=False)
        self.add_parameter("dec", dec, allow_gradient=False)
        self.add_parameter("redshift", redshift, allow_gradient=False)
        self.add_parameter("t0", t0)

        # If the luminosity distance is provided, use that. Otherwise try the
        # redshift value using the cosmology (if given). Finally, default to None.
        if distance is not None:
            self.add_parameter("distance", distance, allow_gradient=False)
        elif redshift is not None and kwargs.get("cosmology") is not None:
            self._redshift_func = RedshiftDistFunc(redshift=self.redshift, **kwargs)
            self.add_parameter("distance", self._redshift_func, allow_gradient=False)
        else:
            self.add_parameter("distance", None, allow_gradient=False)

        # Initialize the effect settings to their default values.
        self.apply_redshift = redshift is not None

        # Set the default effects.
        self.rest_frame_effects = []
        self.obs_frame_effects = []

        # Set the extrapolation for values outside the model's defined bounds.
        self.wave_extrapolation = wave_extrapolation

        # Get a default random number generator for this object, using the
        # given seed if one is provided.
        if seed is None:
            seed = int.from_bytes(urandom(4), "big")
        self._rng = np.random.default_rng(seed=seed)

    def minwave(self):
        """Get the minimum supported wavelength of the model.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        return None

    def maxwave(self):
        """Get the maximum supported wavelength of the model.

        Returns
        -------
        maximum : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        return None

    def set_apply_redshift(self, apply_redshift):
        """Toggles the apply_redshift setting. If set to True, the model will
        apply redshift during the flux density computation including applying wavelength
        and time transformations.

        Parameters
        ----------
        apply_redshift : bool
            The new value for apply_redshift.
        """
        self.apply_redshift = apply_redshift

    def add_effect(self, effect):
        """Add an effect to the model. This effect will be applied to all
        fluxes densities simulated by the model.

        Any effect parameters that are not already in the model
        will be added to this node's parameters.

        Parameters
        ----------
        effect : EffectModel
            The effect to add.
        """
        # Add any effect parameters that are not already in the model.
        for param_name, setter in effect.parameters.items():
            if param_name not in self.setters:
                self.add_parameter(param_name, setter, allow_gradient=False)

        # Add the effect to the appropriate list.
        if effect.rest_frame:
            self.rest_frame_effects.append(effect)
        else:
            self.obs_frame_effects.append(effect)

    def mask_by_time(self, times, graph_state=None):
        """Compute a mask for whether a given time is of interest for a given object.
        For example, a user can use this function to generate a mask to include
        only the observations of interest for a window around the supernova.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        graph_state : GraphState, optional
            An object mapping graph parameters to their values.

        Returns
        -------
        time_mask : numpy.ndarray
            A length T array of Booleans indicating whether the time is of interest.
        """
        return np.full(len(times), True)

    def compute_flux(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free rest frame flux densities.
        The rest-frame flux is defined as F_nu = L_nu / 4*pi*D_L**2,
        where D_L is the luminosity distance.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of rest frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of rest frame SED values (in nJy).
        """
        raise NotImplementedError()

    def compute_flux_with_extrapolation(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object, extrapolating
        to wavelengths where the model is not defined.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        min_query_wave = np.min(wavelengths)
        min_valid_wave = self.minwave()
        if min_valid_wave is None:
            min_valid_wave = min_query_wave

        max_query_wave = np.max(wavelengths)
        max_valid_wave = self.maxwave()
        if max_valid_wave is None:
            max_valid_wave = max_query_wave

        # If no extrapolation is needed, just call compute the flux.
        if min_query_wave >= min_valid_wave and max_query_wave <= max_valid_wave:
            return self.compute_flux(times, wavelengths, graph_state, **kwargs)

        # Truncate the wavelengths on which we evaluate the model.
        before_mask = wavelengths < min_valid_wave
        after_mask = wavelengths > max_valid_wave
        in_range = ~before_mask & ~after_mask

        # Pad the wavelengths with the min and max values and compute the flux at those points.
        query_waves = np.concatenate(([min_valid_wave], wavelengths[in_range], [max_valid_wave]))
        computed_flux = self.compute_flux(times, query_waves, graph_state)

        # Initially zero pad the full array until we fill in the values with extrapolation.
        # We drop the first and last flux values since they were added to get the flux at the
        # min and max wavelengths.
        flux_density = np.zeros((len(times), len(wavelengths)))
        flux_density[:, in_range] = computed_flux[:, 1:-1]

        # Do extrapolation for wavelengths that fell outside the model's bounds.
        if self.wave_extrapolation is not None:
            # Compute the flux values before the model's first valid wavelength.
            flux_density[:, before_mask] = self.wave_extrapolation(
                min_valid_wave,
                computed_flux[:, 0],
                wavelengths[before_mask],
            )

            # Compute the flux values after the model's last valid wavelength.
            flux_density[:, after_mask] = self.wave_extrapolation(
                max_valid_wave,
                computed_flux[:, -1],
                wavelengths[after_mask],
            )

        return flux_density

    def _evaluate_single(self, times, wavelengths, state, rng_info=None, **kwargs):
        """Evaluate the model and apply the effects for a single, given graph state.
        This function applies redshift, computes the flux density for the object,
        applies rest frames effects, performs the redshift correction (if needed),
        and applies the observer frame effects.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.
        """
        if state is None or state.num_samples != 1:
            raise ValueError("A GraphState with num_samples=1 required.")
        params = self.get_local_params(state)

        # Pre-effects are adjustments done to times and/or wavelengths, before flux density
        # computation. We skip if redshift is 0.0 since there is nothing to do.
        if self.apply_redshift and params["redshift"] != 0.0:
            if params.get("redshift", None) is None:
                raise ValueError("The 'redshift' parameter is required for redshifted models.")
            if params.get("t0", None) is None:
                raise ValueError("The 't0' parameter is required for redshifted models.")
            rest_times, rest_wavelengths = obs_to_rest_times_waves(
                times, wavelengths, params["redshift"], params["t0"]
            )
        else:
            rest_times = times
            rest_wavelengths = wavelengths

        # Compute the flux density for the object and apply any rest frame effects.
        flux_density = self.compute_flux_with_extrapolation(rest_times, rest_wavelengths, state, **kwargs)
        for effect in self.rest_frame_effects:
            flux_density = effect.apply(
                flux_density,
                times=rest_times,
                wavelengths=rest_wavelengths,
                rng_info=rng_info,
                **params,
            )

        # Post-effects are adjustments done to the flux density after computation.
        if self.apply_redshift and params["redshift"] != 0.0:
            # We have alread checked that redshift is not None.
            flux_density = rest_to_obs_flux(flux_density, params["redshift"])

        # Apply observer frame effects.
        for effect in self.obs_frame_effects:
            flux_density = effect.apply(
                flux_density,
                times=times,
                wavelengths=wavelengths,
                rng_info=rng_info,
                **params,
            )
        return flux_density

    def evaluate(self, times, wavelengths, graph_state=None, given_args=None, rng_info=None, **kwargs):
        """Draw observations for this object and apply the noise.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState, optional
            An object mapping graph parameters to their values.
        given_args : dict, optional
            A dictionary representing the given arguments for this sample run.
            This can be used as the JAX PyTree for differentiation.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length S x T x N matrix of SED values (in nJy), where S is the number of samples,
            T is the number of time steps, and N is the number of wavelengths.
            If S=1 then the function returns a T x N matrix.
        """
        # Make sure times and wavelengths are numpy arrays.
        times = np.asarray(times)
        wavelengths = np.asarray(wavelengths)

        # Check if we need to sample the graph.
        if graph_state is None:
            graph_state = self.sample_parameters(
                given_args=given_args, num_samples=1, rng_info=rng_info, **kwargs
            )

        results = np.empty((graph_state.num_samples, len(times), len(wavelengths)))
        for sample_num, state in enumerate(graph_state):
            # Compute the flux (applying all effects) and save the result.
            results[sample_num, :, :] = self._evaluate_single(
                times,
                wavelengths,
                state,
                rng_info=rng_info,
                **kwargs,
            )

        if graph_state.num_samples == 1:
            return results[0, :, :]
        return results

    def sample_parameters(self, given_args=None, num_samples=1, rng_info=None):
        """Sample the model's underlying parameters if they are provided by a function
        or ParameterizedModel.

        Parameters
        ----------
        given_args : dict, optional
            A dictionary representing the given arguments for this sample run.
            This can be used as the JAX PyTree for differentiation.
        num_samples : int
            A count of the number of samples to compute.
            Default: 1
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the keyword arguments, including the values needed to sample
            parameters.

        Returns
        -------
        graph_state : GraphState
            An object mapping graph parameters to their values.
        """
        # If the graph has not been sampled ever, update the node positions for
        # every node (model, background, effects).
        if self.node_pos is None:
            self.set_graph_positions()

        graph_state = GraphState(num_samples)
        if given_args is not None:
            graph_state.update(given_args, all_fixed=True)

        # We use the same seen_nodes for all sampling calls so each node
        # is sampled at most one time regardless of link structure.
        seen_nodes = {}
        self._sample_helper(graph_state, seen_nodes, rng_info=rng_info)

        return graph_state

    def get_band_fluxes(self, passband_or_group, times, filters, state, rng_info=None) -> np.ndarray:
        """Get the band fluxes for a given Passband or PassbandGroup.

        Parameters
        ----------
        passband_or_group : Passband or PassbandGroup
            The passband (or passband group) to use.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray or None
            A length T array of filter names. It may be None if
            passband_or_group is a Passband.
        state : GraphState
            An object mapping graph parameters to their values.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        band_fluxes : numpy.ndarray
            A matrix of the band fluxes. If only one sample is provided in the GraphState,
            then returns a length T array. Otherwise returns a size S x T array where S is the
            number of samples in the graph state.
        """
        if isinstance(passband_or_group, Passband):
            if filters is not None and not np.all(filters == passband_or_group.filter_name):
                raise ValueError(
                    "If passband_or_group is a Passband, filters must either be None "
                    "or a list where every entry matches the given filter's name: "
                    f"{passband_or_group.filter_name}."
                )
            spectral_fluxes = self.evaluate(times, passband_or_group.waves, state, rng_info=rng_info)
            return passband_or_group.fluxes_to_bandflux(spectral_fluxes)

        if filters is None:
            raise ValueError("If passband_or_group is a PassbandGroup, filters must be provided.")
        filters = np.asarray(filters)

        band_fluxes = np.empty((state.num_samples, len(times)))
        for filter_name in np.unique(filters):
            passband = passband_or_group[filter_name]
            filter_mask = filters == filter_name
            spectral_fluxes = self.evaluate(times[filter_mask], passband.waves, state, rng_info=rng_info)
            band_fluxes[:, filter_mask] = passband.fluxes_to_bandflux(spectral_fluxes)

        if state.num_samples == 1:
            return band_fluxes[0, :]
        return band_fluxes



import numpy as np


def phot_eff_function(snr):
    """
    Photometric detection efficiency as a simple step function of snr.

    Parameters
    ----------
    snr: list or numpy.ndarray
        Signal to noise ratio of a list of observations.

    Returns
    -------
    eff: list or numpy.ndarray
        The photometric detection efficiency given snr.
    """

    snr = np.array(snr)
    eff = np.where(snr > 5, 1.0, 0.0)

    return eff


def spec_eff_function(peak_imag):
    """
    Spectroscopic follow-up efficiency as a function of peak i band magnitude.

    Parameters
    ----------
    peak_imag: list or numpy.ndarray
        Peak magnitude in i band.

    Returns
    -------
    eff: list or numpy.ndarray
        The spectroscopic efficiency given peak i band magnitude.
        Based on Equation (17) in Kessler et al. 2019
        s0, s1, s2 are fitted using data from Figure 4 in Kessler et al. 2019
    """

    s0 = 1.0
    s1 = 2.36
    s2 = 51.9

    peak_imag = np.array(peak_imag)
    eff = s0 * np.power((1.0 + np.exp(s1 * peak_imag - s2)), -1)

    return eff



import numpy as np
import pytest
from citation_compass import find_in_citations
from sncosmo.models import SALT2Source
from tdastro.sources.salt2_jax import SALT2JaxModel
from tdastro.sources.sncomso_models import SncosmoWrapperModel


def test_salt2_model_parity(test_data_dir):
    """Test loading a SALT2 object from a file and test we get the same
    results as the sncosmo version.
    """
    dir_name = test_data_dir / "truncated-salt2-h17"
    td_model = SALT2JaxModel(x0=0.4, x1=0.3, c=1.1, t0=0.0, model_dir=dir_name)

    # We need to overwrite the source parameter to correspond to
    # the truncated directory data.
    sn_model = SncosmoWrapperModel("SALT2", x0=0.4, x1=0.3, c=1.1, t0=0.0)
    sn_model.source = SALT2Source(modeldir=dir_name)

    # Test compared to values computed via sncosmo's implementation that
    # fall within the range of the truncated grid. We multiple by 1e12
    # for comparison precision purposes.
    times = np.arange(-1.0, 15.0, 0.01)
    waves = np.arange(3800.0, 4200.0, 0.5)

    # Allow TDAstro to return both sets of results in f_nu.
    flux_td = td_model.evaluate(times, waves)
    flux_sn = sn_model.evaluate(times, waves)
    assert np.allclose(flux_td, flux_sn)


def test_salt2_no_model(test_data_dir):
    """Test that we fail if using the wrong model directory."""
    dir_name = test_data_dir / "no_such_salt2_model_dir"
    with pytest.raises(FileNotFoundError):
        _ = SALT2JaxModel(x0=0.5, x1=0.2, c=1.0, t0=0.0, model_dir=dir_name)


def test_salt2_citation():
    """Test the citations for the SALT2 model."""
    salt_citations = find_in_citations("SALT2JaxModel")
    for citation in salt_citations:
        assert "Guy J., 2007" in citation
        assert "sncosmo" in citation



from unittest.mock import patch

from tdastro.utils.data_download import download_data_file_if_needed


def test_download_data_file_if_needed(tmp_path):
    """Test the functionality of downloading a data file using pooch."""
    data_url = "mock"

    def mock_urlretrieve(url, known_hash, fname, path):
        full_name = path / fname
        with open(full_name, "w") as f:
            f.write("Mock data")
        return full_name

    with patch("pooch.retrieve", side_effect=mock_urlretrieve):
        # Create an existing data file.
        data_path_1 = tmp_path / "test_data.dat"
        with open(data_path_1, "w") as f:
            f.write("Test")
        assert data_path_1.exists()

        # If we try to download the file again, it should not overwrite it.
        assert download_data_file_if_needed(data_path_1, data_url, force_download=False)
        with open(data_path_1, "r") as f:
            assert f.read() == "Test"

        # If we force the download, it should overwrite the existing file.
        assert download_data_file_if_needed(data_path_1, data_url, force_download=True)
        with open(data_path_1, "r") as f:
            assert f.read() == "Mock data"

        # Create a second data file that does not exist.
        data_path_2 = tmp_path / "test_data_2.dat"
        assert not data_path_2.exists()

        # Download the second file without forcing it.
        assert download_data_file_if_needed(data_path_2, data_url, force_download=False)
        with open(data_path_2, "r") as f:
            assert f.read() == "Mock data"



import astropy.constants as const
import astropy.units as u
import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.spatial.transform import Rotation
from tdastro.sources.periodic_variable_star import EclipsingBinaryStar


def test_circle_overlap_area_total_eclipse():
    """Test that for r1=r2 and d=0, the overlap area is equal to the area of the circles"""
    r = 3.5
    d = 0.0
    overlap_area = EclipsingBinaryStar._circle_overlap_area(d, r, r)
    assert_allclose(overlap_area, np.pi * r**2)


def test_circle_overlap_area_primary_minimum():
    """Test that for r1>r2 and d < r1-r2, the overlap area is equal to the area of the secondary circle"""
    r1 = 3.5
    r2 = 1.5
    d = 1.0
    overlap_area = EclipsingBinaryStar._circle_overlap_area(d, r1, r2)
    assert_allclose(overlap_area, np.pi * r2**2)


def test_circle_overlap_area_no_overlap():
    """Test that for d>=r1+r2, the overlap area is zero"""
    r1 = 3.5
    r2 = 1.5
    d = r1 + r2
    overlap_area = EclipsingBinaryStar._circle_overlap_area(d, r1, r2)
    assert_allclose(overlap_area, 0.0)


@pytest.mark.parametrize(
    "d, r1, r2",
    [
        # Circle centers are inside circles.
        (1.0, 3.5, 3.0),
        # Intersections and secondary center lie on the same line.
        (3.0, 5.0, 4.0),
        # Center of the secondary is still inside the primary, but near the edge.
        (4.5, 5.0, 4.0),
        # Circle centers are outside each other.
        (5.0, 3.5, 3.0),
    ],
)
def test_circle_overlap_area_with_monte_carlo(d, r1, r2):
    """Test EclipsingBinaryStar._circle_overlap_area with MC simulation"""
    # The origin is the center of the primary star, the secondary star is at (d, 0).

    n_samples = 1_000_000
    rng = np.random.default_rng(0)

    # Generate random points in the bounding box
    x_range = [-r1, d + r2]
    assert r1 >= r2, "Check the test parameters, r1 should be not less than r2."
    y_range = [-r1, r1]
    x, y = rng.uniform([x_range[0], y_range[0]], [x_range[1], y_range[1]], (n_samples, 2)).T

    # Check if the points are inside the circles and find the overlap area
    inside_primary = np.hypot(x, y) < r1
    inside_secondary = np.hypot(x - d, y) < r2
    n_overlap = np.sum(inside_primary & inside_secondary)
    desired_overlap_area = n_overlap / n_samples * (x_range[1] - x_range[0]) * (y_range[1] - y_range[0])

    actual_overlap_area = EclipsingBinaryStar._circle_overlap_area(d, r1, r2)
    # 2-sigma tolerance
    assert_allclose(actual_overlap_area, desired_overlap_area, rtol=2 / np.sqrt(n_overlap))


def test_norm_star_center_distance_zero():
    """Test EclipsingBinaryStar._norm_star_center_distance for phase_fraction=0.0, inclination_degree=90.0"""
    assert_allclose(EclipsingBinaryStar._norm_star_center_distance(0.0, 90.0), 0.0, atol=1e-12)


@pytest.mark.parametrize(
    "phase_fraction, inclination_degree",
    [
        # Inclination is 0, the distance is always 1
        (0.0, 0.0),
        (0.25, 0.0),
        (0.3, 0.0),
        (0.5, 0.0),
        (0.75, 0.0),
        (0.95, 0.0),
        # Phase is 0.25 or 0.75, the distance is always 1
        (0.25, 0.0),
        (0.75, 10.0),
        (0.25, 45.0),
        (0.75, 57.0),
        (0.25, 85.0),
    ],
)
def test_norm_star_center_distance_one(phase_fraction, inclination_degree):
    """Test EclipsingBinaryStar._norm_star_center_distance returns 1.0 for maximum separation"""
    assert_allclose(EclipsingBinaryStar._norm_star_center_distance(phase_fraction, inclination_degree), 1.0)


def test_norm_star_center_distance_with_spherical_geometry():
    """Test EclipsingBinaryStar._norm_star_center_distance with spherical geometry implementation"""
    # We don't need reproducible random numbers here, because test should work for any random input.
    n_samples = 100
    phase = np.random.uniform(0, 1, n_samples)
    inclination = np.random.uniform(0, 90, n_samples)

    phase_radians = 2 * np.pi * phase
    # Let secondary orbit define x' y' plane, then coordinates are:
    x_prime, y_prime, z_prime = np.cos(phase_radians), np.sin(phase_radians), np.zeros_like(phase_radians)
    # Rotate x' y' z' to x y z plane with matrix multiplication.
    # Angle between z and z' is inclination, y is y', so we rotate about y axis.
    rotation = Rotation.from_euler("y", 90.0 - inclination, degrees=True)
    x, y, z = rotation.apply(np.stack([x_prime, y_prime, z_prime], axis=-1)).T
    # x is looking to the observer, y and z are the observer's plane.
    expected_distance = np.hypot(y, z)
    actual_distance = EclipsingBinaryStar._norm_star_center_distance(phase, inclination)
    assert_allclose(actual_distance, expected_distance)


def test_eclipsing_binary_star():
    """Test EclipsingBinaryStar light curve satisfies astrophysical expectations"""
    points_per_period = 1000

    distance_pc = 10
    primary_mass = 1.0 * u.M_sun
    secondary_mass = 0.6 * u.M_sun
    major_semiaxis = 10 * u.R_sun

    # Mass-radius relation for main-sequence stars
    primary_radius = 1.0 * u.R_sun * (primary_mass / const.M_sun) ** 0.8
    secondary_radius = 1.0 * u.R_sun * (secondary_mass / const.M_sun) ** 0.8
    assert primary_radius > secondary_radius, "Primary should be larger than secondary"
    # Luminosity-mass relation for main-sequence stars
    primary_lum = 1.0 * u.L_sun * (primary_mass / const.M_sun) ** 4
    secondary_lum = 1.0 * u.L_sun * (secondary_mass / const.M_sun) ** 4
    # Effective temperature from Stefan-Boltzmann law
    primary_temperature = (primary_lum / (4 * np.pi * primary_radius**2 * const.sigma_sb)) ** 0.25
    secondary_temperature = (secondary_lum / (4 * np.pi * secondary_radius**2 * const.sigma_sb)) ** 0.25
    assert primary_temperature > secondary_temperature, "Primary should be hotter than secondary"

    # Third law of Kepler, with gravitation constant
    period = np.sqrt(4 * np.pi**2 * major_semiaxis**3 / (const.G * (primary_mass + secondary_mass)))

    source = EclipsingBinaryStar(
        distance=distance_pc,
        period=period.to_value(u.day),
        t0=0.0,
        major_semiaxis=major_semiaxis.cgs.value,
        inclination=89.0,
        primary_radius=primary_radius.cgs.value,
        primary_temperature=primary_temperature.cgs.value,
        secondary_radius=secondary_radius.cgs.value,
        secondary_temperature=secondary_temperature.cgs.value,
    )
    # Times in days
    times = np.linspace(0, 2.0 * period.to_value(u.day), 2 * points_per_period + 1)
    # Wavelengths in cm
    wavelengths_aa = np.array([4500, 6000])

    fluxes = source.evaluate(times, wavelengths_aa)

    # Check the fluxes are positive
    assert np.all(fluxes >= 0)
    # Check the fluxes are periodic
    assert_allclose(fluxes[0], fluxes[-1])
    assert_allclose(fluxes[:points_per_period], fluxes[points_per_period : 2 * points_per_period])
    # Check the fluxes are symmetric
    assert_allclose(fluxes[: points_per_period + 1], fluxes[points_per_period::-1])

    # Check the primary minimum is deeper than everything else
    assert np.all(fluxes[0] <= fluxes)

    # Check if primary minimum is redder than everything else
    color = -2.5 * np.log10(fluxes[:, 0] / fluxes[:, 1])
    assert np.all(color[0] >= color)



import numpy as np
from tdastro.astro_utils.passbands import PassbandGroup
from tdastro.math_nodes.given_sampler import GivenValueList
from tdastro.opsim.opsim import OpSim
from tdastro.simulate import simulate_lightcurves
from tdastro.sources.basic_sources import StaticSource


def test_simulate_lightcurves(test_data_dir):
    """Test an end to end run of simulating the lightcurves."""
    # Load the OpSim data.
    opsim_db = OpSim.from_db(test_data_dir / "opsim_small.db")

    # Load the passband data for the griz filters only.
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=test_data_dir / "passbands",
        filters=["g", "r", "i", "z"],
    )

    # Create a static source with known brightnesses and RA, dec
    # values that match the opsim.
    given_brightness = [1000.0, 2000.0, 5000.0, 1000.0, 100.0]
    source = StaticSource(
        brightness=GivenValueList(given_brightness),
        t0=0.0,
        ra=GivenValueList(opsim_db["ra"].values[0:5]),
        dec=GivenValueList(opsim_db["dec"].values[0:5]),
        redshift=0.0,
        node_label="source",
    )

    results = simulate_lightcurves(
        source,
        5,
        opsim_db,
        passband_group,
        opsim_save_cols=["observationId", "zp_nJy"],
        param_cols=["source.brightness"],
    )
    assert len(results) == 5
    assert np.all(results["nobs"].values >= 1)
    for idx in range(5):
        num_obs = results["nobs"][idx]
        assert len(results.loc[idx]["lightcurve"]["flux"]) == num_obs

        # Check that we pulled the metadata from the opsim.
        assert len(results.loc[idx]["lightcurve"]["observationId"]) == num_obs
        assert len(np.unique(results.loc[idx]["lightcurve"]["observationId"])) == num_obs
        assert np.all(results.loc[idx]["lightcurve"]["observationId"] >= 0)
        assert len(results.loc[idx]["lightcurve"]["zp_nJy"]) == num_obs

        # Check that we extract one of the parameters.
        assert results["source_brightness"][idx] == given_brightness[idx]



from abc import ABC, abstractmethod

import numpy as np

from tdastro.astro_utils.black_body import black_body_luminosity_density_per_solid
from tdastro.consts import ANGSTROM_TO_CM, CGS_FNU_UNIT_TO_NJY, PARSEC_TO_CM
from tdastro.sources.periodic_source import PeriodicSource


class PeriodicVariableStar(PeriodicSource, ABC):
    """A model for a periodic variable star.

    Parameterized values include:
      * dec - The object's declination in degrees. [from PhysicalModel]
      * distance - The object's luminosity distance in pc. [from PhysicalModel]
      * period - The period of the source, in days. [from PeriodicSource]
      * ra - The object's right ascension in degrees. [from PhysicalModel]
      * redshift - The object's redshift. [from PhysicalModel]
      * t0 - The t0 of the zero phase, date. [from PhysicalModel]

    Attributes
    ----------
    period : float
        The period of the source, in days.
    t0 : float
        The t0 of the zero phase, date. Could be date of the minimum or maximum light
        or any other reference time point.
    distance : float
        The distance to the source, in pc.
    """

    def __init__(self, period, **kwargs):
        super().__init__(period, **kwargs)
        if not self.has_valid_param("distance"):
            raise ValueError("Distance parameter is required for PeriodicVariableStar")

    def _evaluate_phases(self, phases, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object, as a function of phase.

        Parameters
        ----------
        phases : numpy.ndarray
            A length T array of phases, in the range [0, 1].
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths, in angstroms.
        graph_state : dict, optional
            A given setting of all the parameters and their values.
        **kwargs : dict, optional
              Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values, in nJy.
        """
        distance_cm = self.get_param(graph_state, "distance") * PARSEC_TO_CM
        wavelengths_cm = wavelengths * ANGSTROM_TO_CM
        dl_dnu_domega_cgs = self._dl_dnu_domega_phases(phases, wavelengths_cm, graph_state, **kwargs)
        return dl_dnu_domega_cgs / np.square(distance_cm) * CGS_FNU_UNIT_TO_NJY

    @abstractmethod
    def _dl_dnu_domega_phases(self, phases, wavelengths_cm, graph_state, **kwargs):
        r"""Draw effect-free luminosity density per unit solid angle, as a function of phase.

        It is dL / d \nu / d \Omega, so the units are erg / s / Hz / sr.
        Sometimes it is called "observed luminosity density".
        L_nu = \int this_value d \Omega.
        For isotropic source L_nu = 4 \pi \int this_value d \Omega.
        Bolometric luminosity is \int this_value d \nu d \Omega.
        Flux density (static Universe) F_nu = this_value / distance^2.

        Parameters
        ----------
        phases : numpy.ndarray
            A length T array of phases, in the range [0, 1].
        wavelengths_cm : numpy.ndarray, optional
            A length N array of wavelengths in cm.
        graph_state : dict, optional
            A given setting of all the parameters and their values.
        **kwargs : dict, optional
              Any additional keyword arguments.

        Returns
        -------
        luminosity_density_per_solid_angle : numpy.ndarray
            A length T x N matrix of luminosity density per unit solid angle values.
            Units are CGS, erg/s/Hz/steradian.
        """
        raise NotImplementedError()


class EclipsingBinaryStar(PeriodicVariableStar):
    """A toy model for a detached eclipsing binary star.

    It is assumed that the stars are spherical, SED is black-body,
    and the orbits are circular. t0 is the epoch of the primary minimum.
    No limb darkening, reflection, or other effects are included.

    Parameterized values include:
      * dec - The object's declination in degrees. [from PhysicalModel]
      * distance - The object's luminosity distance in pc. [from PhysicalModel]
      * inclination - The inclination of the orbit, in degrees.
      * major_semiaxis - The major semiaxis of the orbit, in AU.
      * period - The period of the source, in days. [from PeriodicSource]
      * primary_radius - The radius of the primary star, in solar radii.
      * primary_temperature - The effective temperature of the primary star, in kelvins.
      * ra - The object's right ascension in degrees. [from PhysicalModel]
      * redshift - The object's redshift. [from PhysicalModel]
      * secondary_radius - The radius of the secondary star, in solar radii.
      * secondary_temperature - The effective temperature of the secondary star, in kelvins.
      * t0 - The t0 of the zero phase, date. [from PeriodicSource]

    Attributes
    ----------
    period : float
        The period of the source, in days.
    major_semiaxis : float
        The major semiaxis of the orbit, in AU.
    inclination : float
        The inclination of the orbit, in degrees.
    primary_radius : float
        The radius of the primary star, in solar radii.
    secondary_radius : float
        The radius of the secondary star, in solar radii.
    primary_temperature : float
        The effective temperature of the primary star, in kelvins.
    secondary_temperature : float
        The effective temperature of the secondary star, in kelvins.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter("major_semiaxis", **kwargs)
        self.add_parameter("inclination", **kwargs)
        self.add_parameter("primary_radius", **kwargs)
        self.add_parameter("secondary_radius", **kwargs)
        self.add_parameter("primary_temperature", **kwargs)
        self.add_parameter("secondary_temperature", **kwargs)

    def _dl_dnu_domega_phases(self, phases, wavelengths_cm, graph_state, **kwargs):
        """Draw effect-free luminosity density for this object, as a function of phase.

        Parameters
        ----------
        phases : numpy.ndarray
            A length T array of phases, in the range [0, 1].
        wavelengths_cm : numpy.ndarray, optional
            A length N array of wavelengths, in cm.
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
              Any additional keyword arguments.

        Returns
        -------
        luminosity_density : numpy.ndarray
            A length T x N matrix of luminosity density values.
            Output is in CGS units of erg/s/Hz/steradian.
        """
        phases = phases[:, None]

        # Extract the parameters we need.
        params = self.get_local_params(graph_state)
        primary_temperature = params["primary_temperature"]
        primary_radius = params["primary_radius"]
        secondary_temperature = params["secondary_temperature"]
        secondary_radius = params["secondary_radius"]
        major_semiaxis = params["major_semiaxis"]
        inclination = params["inclination"]

        primary_lum = black_body_luminosity_density_per_solid(
            primary_temperature, primary_radius, wavelengths_cm
        )
        secondary_lum = black_body_luminosity_density_per_solid(
            secondary_temperature, secondary_radius, wavelengths_cm
        )

        # Distance between the centers of the stars on the plane of the sky
        star_centers_distance = major_semiaxis * self._norm_star_center_distance(phases, inclination)

        overlap_area = self._circle_overlap_area(star_centers_distance, primary_radius, secondary_radius)

        # For phases around the main minimum (0) primary star is eclipsed
        secondary_star_closer = np.logical_or(phases < 0.25, phases > 0.75)
        primary_star_overlap_ratio = np.where(
            secondary_star_closer, overlap_area / (np.pi * primary_radius**2), 0.0
        )
        secondary_star_overlap_ratio = np.where(
            ~secondary_star_closer, overlap_area / (np.pi * secondary_radius**2), 0.0
        )

        primary_lum_eclipsed = primary_lum * (1.0 - primary_star_overlap_ratio)
        secondary_lum_eclipsed = secondary_lum * (1.0 - secondary_star_overlap_ratio)

        # Just in case, we clip negative values to zero
        total_lum = np.where(primary_lum_eclipsed >= 0, primary_lum_eclipsed, 0) + np.where(
            secondary_lum_eclipsed >= 0, secondary_lum_eclipsed, 0
        )

        return total_lum

    @staticmethod
    def _norm_star_center_distance(phase_fraction, inclination_degree):
        """Calculate the distance between the centers of the stars on the observer's plane.

        Parameters
        ----------
        phase_fraction : np.ndarray
            The phase of the orbit, in the range [0, 1].
        inclination_degree : float
            The inclination of the orbit, in degrees.

        Returns
        -------
        distance : np.ndarray
            The distance between the centers of the stars on the plane of the sky,
            normalized by the major semiaxis.
        """
        phase_radians = 2 * np.pi * phase_fraction
        inclination_radians = np.radians(inclination_degree)
        return np.hypot(np.cos(inclination_radians) * np.cos(phase_radians), np.sin(phase_radians))

    @staticmethod
    def _circle_overlap_area(d, r1, r2):
        """Calculate the area of overlap between two circles.

        Parameters
        ----------
        d : np.ndarray
            The distance between the centers of the circles.
        r1 : float
            The radius of the first circle.
        r2 : float
            The radius of the second circle.

        Returns
        -------
        overlap_area : np.ndarray
            The area of overlap between the two circles.
        """

        # We consider four points here: the centers of the circles and the points where the circles intersect
        # The intersection region is a union of two segments of the circles, while their intersection is
        # a quadrilateral. So the area of the intersection is the sum of the areas of the two segments minus
        # the area of the quadrilateral. The intersection region is symmetric over the line connecting
        # the centers of the circles, so we can consider only one half of it, and multiply the result by 2.
        # In this half region intersection is a triangle with sides d, r1, and r2.

        r1_sq = np.square(r1)
        r2_sq = np.square(r2)
        d_sq = np.square(d)

        # We mute warnings for division by zero and invalid values in arccos, as we handle them manually
        with np.errstate(divide="ignore", invalid="ignore"):
            # Angles between lines connecting the centers of the circles and the points of intersection.
            # These are halves of the intersection-center-intersection angles.
            alpha1 = np.arccos((d_sq + r1_sq - r2_sq) / (2 * r1 * d))
            alpha2 = np.arccos((d_sq + r2_sq - r1_sq) / (2 * r2 * d))

            # Area of circular segments, it is 1/2 r^2 angle, angle = 2 alpha, so area = r^2 alpha
            area1 = r1_sq * alpha1
            area2 = r2_sq * alpha2

            # Area of the intersection quadrilateral, twice triangular area which is 1/2 r d sin(alpha).
            # Should be symmetric over 1-2 index swap.
            triangle_area = r1 * d * np.sin(alpha1)

        area = area1 + area2 - triangle_area

        # Just in case, we clip negative values to zero
        area = np.where(area >= 0.0, area, 0.0)

        # Account for the total overlap
        area = np.where(d <= np.abs(r1 - r2), np.pi * np.square(np.minimum(r1, r2)), area)
        # Account for the no overlap
        area = np.where(d >= r1 + r2, 0, area)

        return area



from unittest.mock import patch

import numpy as np
import pytest
from sncosmo import Bandpass
from tdastro.astro_utils.passbands import Passband, PassbandGroup
from tdastro.sources.spline_model import SplineModel


def create_lsst_passband_group(passbands_dir, delta_wave=5.0, trim_quantile=None):
    """Helper function to create a PassbandGroup object for LSST passbands, using transmission tables
    included in the test data directory."""
    return PassbandGroup.from_preset(
        preset="LSST",
        table_dir=passbands_dir,
        delta_wave=delta_wave,
        trim_quantile=trim_quantile,
    )


def create_toy_passband_group(path, delta_wave=5.0, trim_quantile=None):
    """Helper function to create a toy PassbandGroup object for testing.

    Notable features of this toy passband group:
    - Three passbands: A, B, C
    - Each transmission table includes only three rows
    - Passbands A and B have overlapping wavelength ranges
    - Passbands A and C are defined with a 100 unit grid step, while passband B is defined with a 50 unit step
    """
    # Initialize requirements for PassbandGroup object
    survey = "TOY"
    filter_names = ["a", "b", "c"]
    table_dir = path / survey
    table_dir.mkdir()

    transmission_tables = {
        "a": "100 0.5\n200 0.75\n300 0.25\n",
        "b": "250 0.25\n300 0.5\n350 0.75\n",
        "c": "400 0.75\n500 0.25\n600 0.5\n",
    }
    for filter_name in filter_names:
        with open(table_dir / f"{filter_name}.dat", "w") as f:
            f.write(transmission_tables[filter_name])

    # Load the PassbandGroup object
    passbands = []
    for filter_name in filter_names:
        passbands.append(
            {
                "survey": survey,
                "filter_name": filter_name,
                "table_path": table_dir / f"{filter_name}.dat",
                "delta_wave": delta_wave,
                "trim_quantile": trim_quantile,
            }
        )
    return PassbandGroup(passbands)


def test_passband_group_access(tmp_path):
    """Test that we can create a passband group and access the individual passbands."""
    table_vals = np.array([[100, 0.5], [200, 0.75], [300, 0.25]])
    pb_list = [
        Passband(table_vals, "survey1", "a", trim_quantile=None),
        Passband(table_vals, "survey1", "b", trim_quantile=None),
        Passband(table_vals, "survey1", "c", trim_quantile=None),
        Passband(table_vals, "survey2", "c", trim_quantile=None),
        Passband(table_vals, "survey2", "d", trim_quantile=None),
    ]

    pb_group = PassbandGroup(given_passbands=pb_list)
    assert len(pb_group) == 5
    assert "survey1_a" in pb_group
    assert "survey1_b" in pb_group
    assert "survey1_c" in pb_group
    assert "survey2_c" in pb_group
    assert "survey2_d" in pb_group

    # We can access the passbands by their full name.
    assert pb_group["survey1_a"].filter_name == "a"
    assert pb_group["survey1_b"].filter_name == "b"
    assert pb_group["survey1_c"].filter_name == "c"
    assert pb_group["survey2_c"].filter_name == "c"
    assert pb_group["survey2_d"].filter_name == "d"

    # Test that we can access a list of filters.
    assert pb_group.filters == ["a", "b", "c", "d"]

    # If the group only has a single passband for a given filter,
    # We can use the filter name to access it.
    assert pb_group["a"].full_name == "survey1_a"
    assert pb_group["b"].full_name == "survey1_b"
    assert pb_group["d"].full_name == "survey2_d"

    # But we get an error if we try to look up by a filter name that occurs twice.
    with pytest.raises(KeyError):
        _ = pb_group["c"]

    # The contains functionality will work with all filter names.
    assert "a" in pb_group
    assert "b" in pb_group
    assert "c" in pb_group
    assert "d" in pb_group

    # Check that we can filter a list of filter_names by whether they occur in the passband group.
    filters = ["a", "b", "e", "f", "c", "1", "2", "d", "a", "a"]
    expected = [True, True, False, False, True, False, False, True, True, True]
    assert np.array_equal(pb_group.mask_by_filter(filters), expected)


def test_passband_group_init(tmp_path, passbands_dir):
    """Test the initialization of the Passband class, and implicitly, _load_preset."""
    # Test that we cannot create an empty PassbandGroup object
    with pytest.raises(TypeError):
        _ = PassbandGroup()

    # Test that the PassbandGroup class raises an error for an unknown preset
    try:
        _ = PassbandGroup.from_preset(preset="Unknown")
    except ValueError as e:
        assert str(e) == "Unknown passband preset: Unknown"
    else:
        raise AssertionError("PassbandGroup should raise an error for an unknown preset")

    # Test that the PassbandGroup class can be initialized with a preset
    lsst_passband_group = create_lsst_passband_group(passbands_dir)
    assert len(lsst_passband_group.passbands) == 6
    assert len(lsst_passband_group) == 6
    assert "LSST_u" in lsst_passband_group
    assert "LSST_g" in lsst_passband_group
    assert "LSST_r" in lsst_passband_group
    assert "LSST_i" in lsst_passband_group
    assert "LSST_z" in lsst_passband_group
    assert "LSST_y" in lsst_passband_group
    assert "LSST_purple" not in lsst_passband_group

    # Test that we fail if we use the wrong units for the LSST preset
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_preset(
            preset="LSST",
            table_dir=passbands_dir,
            delta_wave=5.0,
            trim_quantile=None,
            units="A",
        )

    # We can access passbands using the [] notation.
    assert lsst_passband_group["LSST_u"].filter_name == "u"
    with pytest.raises(KeyError):
        _ = lsst_passband_group["LSST_purple"]

    # Test that the PassbandGroup class can be initialized with a dict of passband parameters
    lsst_gri_passband_parameters = [
        {"survey": "LSST", "filter_name": "g", "table_path": f"{passbands_dir}/LSST/g.dat"},
        {"survey": "LSST", "filter_name": "r", "table_path": f"{passbands_dir}/LSST/r.dat"},
        {"survey": "LSST", "filter_name": "i", "table_path": f"{passbands_dir}/LSST/i.dat"},
    ]
    lsst_gri_passband_group = PassbandGroup(given_passbands=lsst_gri_passband_parameters)
    assert len(lsst_gri_passband_group) == 3
    assert "LSST_g" in lsst_gri_passband_group
    assert "LSST_r" in lsst_gri_passband_group
    assert "LSST_i" in lsst_gri_passband_group

    # Test our toy passband group, which makes a PassbandGroup using a custom passband parameters dictionary
    toy_passband_group = create_toy_passband_group(tmp_path)
    assert len(toy_passband_group) == 3
    assert "TOY_a" in toy_passband_group
    assert "TOY_b" in toy_passband_group
    assert "TOY_c" in toy_passband_group
    np.testing.assert_allclose(
        toy_passband_group.waves,
        np.unique(np.concatenate([np.arange(100, 301, 5), np.arange(250, 351, 5), np.arange(400, 601, 5)])),
    )

    # Test that we can retrieve bounds for this passband group.
    min_w, max_w = toy_passband_group.wave_bounds()
    assert min_w == 100.0
    assert max_w == 600.0

    # Test that the PassbandGroup class raises an error for an unknown preset
    try:
        _ = PassbandGroup.from_preset(preset="Unknown")
    except ValueError as e:
        assert str(e) == "Unknown passband preset: Unknown"
    else:
        raise AssertionError("PassbandGroup should raise an error for an unknown preset")


def test_passband_group_from_dir(tmp_path):
    """Test that we can load a PassbandGroup from a directory."""
    survey = "FAKE"

    # Create a new survey directory and fill it with filter files.
    table_dir = tmp_path / survey
    table_dir.mkdir()
    transmission_tables = {
        "a": "100 0.5\n150 0.75\n200 0.25\n",
        "b": "200 0.25\n250 0.5\n300 0.75\n",
        "c": "300 0.75\n350 0.25\n400 0.5\n",
        "d": "400 0.75\n450 0.25\n500 0.5\n",
        "e": "500 0.75\n550 0.25\n600 0.5\n",
    }
    for filter_name in transmission_tables:
        with open(table_dir / f"{filter_name}.dat", "w") as f:
            f.write(transmission_tables[filter_name])

    # Load four of the filters from the PassbandGroup
    load_filters = ["a", "b", "c", "e"]
    pb_group = PassbandGroup.from_dir(table_dir, filters=load_filters)
    assert len(pb_group) == len(load_filters)

    for filter in load_filters:
        assert filter in pb_group

        wave_start = int(transmission_tables[filter].split()[0])
        assert wave_start == pb_group[filter]._loaded_table[0][0]

    # Check that we throw an error if we try to access an invalid directory.
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_dir("./no_such_directory", filters=["a", "b"])

    # Check that we throw an error if we try to load a filter that does not exist.
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_dir(table_dir, filters=["a", "b", "z"])


def test_passband_group_from_list(tmp_path):
    """Test that we can create a PassbandGroup from a pre-specified list."""
    pb_list = [
        Passband(
            np.array([[100, 0.5], [200, 0.75], [300, 0.25]]),
            "my_survey",
            "a",
            trim_quantile=None,
        ),
        Passband(
            np.array([[250, 0.25], [300, 0.5], [350, 0.75]]),
            "my_survey",
            "b",
            trim_quantile=None,
        ),
        Passband(
            np.array([[400, 0.75], [500, 0.25], [600, 0.5]]),
            "my_survey",
            "c",
            trim_quantile=None,
        ),
    ]
    test_passband_group = PassbandGroup(given_passbands=pb_list)
    assert len(test_passband_group) == 3
    assert "my_survey_a" in test_passband_group
    assert "my_survey_b" in test_passband_group
    assert "my_survey_c" in test_passband_group

    all_waves = np.concatenate([np.arange(100, 301, 5), np.arange(250, 351, 5), np.arange(400, 601, 5)])
    assert np.allclose(test_passband_group.waves, np.unique(all_waves))

    # Test that we can manually add another passband to the group.
    pb4 = Passband(
        np.array([[800, 0.75], [850, 0.25], [900, 0.5]]),
        "my_survey2",
        "a",
        trim_quantile=None,
    )
    test_passband_group.add_passband(pb4)

    assert len(test_passband_group) == 4
    assert "my_survey_a" in test_passband_group
    assert "my_survey_b" in test_passband_group
    assert "my_survey_c" in test_passband_group
    assert "my_survey2_a" in test_passband_group

    all_waves = np.concatenate([all_waves, np.arange(800, 901, 5)])
    assert np.allclose(test_passband_group.waves, np.unique(all_waves))


def test_passband_load_subset_passbands(tmp_path):
    """Test that we can load a subset of filters."""
    pb_list = [
        Passband(
            np.array([[100, 0.5], [200, 0.75], [300, 0.25]]),
            "my_survey",
            "a",
            trim_quantile=None,
        ),
        Passband(
            np.array([[250, 0.25], [300, 0.5], [350, 0.75]]),
            "my_survey",
            "b",
            trim_quantile=None,
        ),
        Passband(
            np.array([[400, 0.75], [500, 0.25], [600, 0.5]]),
            "my_survey",
            "c",
            trim_quantile=None,
        ),
        Passband(
            np.array([[800, 0.75], [850, 0.25], [900, 0.5]]),
            "my_survey",
            "d",
            trim_quantile=None,
        ),
    ]

    # Load one filter by full name and the other by filter name.
    test_passband_group = PassbandGroup(given_passbands=pb_list, filters=["my_survey_a", "c"])
    assert len(test_passband_group) == 2

    assert np.allclose(
        test_passband_group.waves,
        np.unique(np.concatenate([np.arange(100, 301, 5), np.arange(400, 601, 5)])),
    )

    # We run into an error if we try to load a filter that does not exist.
    with pytest.raises(ValueError):
        _ = PassbandGroup(given_passbands=pb_list, filters=["my_survey_a", "z"])


def test_passband_ztf_preset():
    """Test that we can load the ZTF passbands."""

    def mock_get_bandpass(name):
        """Return a predefined Bandpass object instead of downloading the transmission table."""
        return Bandpass(np.array([6000, 6005, 6010]), np.array([0.5, 0.6, 0.7]))

    # Mock the get_bandpass portion of the download method
    with patch("sncosmo.get_bandpass", side_effect=mock_get_bandpass):
        group = PassbandGroup.from_preset(preset="ZTF")
        assert len(group) == 3
        assert "ZTF_g" in group
        assert "ZTF_r" in group
        assert "ZTF_i" in group

    # Try the load with a subset of filters.
    with patch("sncosmo.get_bandpass", side_effect=mock_get_bandpass):
        group = PassbandGroup.from_preset(preset="ZTF", filters=["g", "i"])
        assert len(group) == 2
        assert "ZTF_g" in group
        assert "ZTF_i" in group
        assert "ZTF_r" not in group


def test_passband_invalid_preset():
    """Test that we throw an error when given an invalid preset name."""
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_preset(preset="Invalid")


def test_passband_unique_waves():
    """Test that if we create two passbands with very similar wavelengths, they get merged."""
    pb_list = [
        Passband(
            np.array([[100, 0.5], [250, 0.25]]),
            "my_survey",
            "a",
            trim_quantile=None,
        ),
        Passband(
            np.array([[200.000001, 0.25], [300.000001, 0.5]]),
            "my_survey",
            "b",
            trim_quantile=None,
        ),
    ]
    test_passband_group = PassbandGroup(given_passbands=pb_list)
    assert np.allclose(
        test_passband_group.waves,
        np.concatenate([np.arange(100, 201, 5), np.arange(205.000001, 301, 5)]),
    )

    # Larger gaps won't register.
    pb_list = [
        Passband(
            np.array([[100, 0.5], [250, 0.25]]),
            "my_survey",
            "a",
            trim_quantile=None,
        ),
        Passband(
            np.array([[200.5, 0.25], [300.5, 0.5]]),
            "my_survey",
            "b",
            trim_quantile=None,
        ),
    ]
    test_passband_group = PassbandGroup(given_passbands=pb_list)
    assert np.allclose(
        test_passband_group.waves,
        np.unique(np.concatenate([np.arange(100, 251, 5), np.arange(200.5, 301, 5)])),
    )


def test_passband_group_str(passbands_dir, tmp_path):
    """Test the __str__ method of the PassbandGroup class."""
    toy_passband_group = create_toy_passband_group(tmp_path)
    assert str(toy_passband_group) == "PassbandGroup containing 3 passbands: TOY_a, TOY_b, TOY_c"

    lsst_passband_group = create_lsst_passband_group(passbands_dir)
    assert str(lsst_passband_group) == (
        "PassbandGroup containing 6 passbands: LSST_u, LSST_g, LSST_r, LSST_i, LSST_z, LSST_y"
    )


def test_passband_group_fluxes_to_bandfluxes(passbands_dir):
    """Test the fluxes_to_bandfluxes method of the PassbandGroup class."""
    # Test with simple passband group
    lsst_passband_group = create_lsst_passband_group(passbands_dir, delta_wave=20, trim_quantile=None)

    # Create some mock flux data with the same number of columns as wavelengths in our PassbandGroup
    flux = np.linspace(lsst_passband_group.waves * 10, 100, 5)
    bandfluxes = lsst_passband_group.fluxes_to_bandfluxes(flux)

    # Check structure of result is as expected (see Passband tests for value tests)
    assert len(bandfluxes) == 6
    for band_name in lsst_passband_group.passbands:
        assert band_name in bandfluxes
        assert bandfluxes[band_name].shape == (5,)

        # Check that we get the same result with fluxes_to_bandflux.
        bandflux = lsst_passband_group.fluxes_to_bandflux(flux, band_name)
        np.testing.assert_allclose(bandflux, bandfluxes[band_name])


def test_passband_group_wrapped_from_physical_source(passbands_dir, tmp_path):
    """Test get_band_fluxes, PhysicalModel's wrapped version of PassbandGroup's fluxes_to_bandfluxes."""
    times = np.array([1.0, 2.0, 3.0])
    wavelengths = np.array([10.0, 20.0, 30.0])
    fluxes = np.array([[1.0, 5.0, 1.0], [5.0, 10.0, 5.0], [1.0, 5.0, 3.0]])
    model = SplineModel(times, wavelengths, fluxes, time_degree=1, wave_degree=1)
    state = model.sample_parameters()

    test_times = np.array([1.0, 1.5, 2.0, 2.5, 3.0, 3.5])

    # Test wrapper with a PassbandGroup as input (see Passband tests for single-band tests)
    # Using LSST passband group:
    lsst_passband_group = create_lsst_passband_group(passbands_dir, delta_wave=20, trim_quantile=None)
    n_lsst_bands = len(lsst_passband_group.passbands)
    n_times = len(test_times)

    fluxes_source_model = model.get_band_fluxes(
        lsst_passband_group,
        times=np.repeat(test_times, n_lsst_bands),
        filters=np.tile(list(lsst_passband_group.passbands.keys()), n_times),
        state=state,
    )
    result_from_source_model = {
        filter_name: fluxes_source_model[i::n_lsst_bands]
        for i, filter_name in enumerate(lsst_passband_group.passbands)
    }
    evaluated_fluxes = model.evaluate(test_times, lsst_passband_group.waves, state)
    result_from_passband_group = lsst_passband_group.fluxes_to_bandfluxes(evaluated_fluxes)

    # Check the two dicts are the same
    assert result_from_source_model.keys() == result_from_passband_group.keys()
    for key in result_from_source_model:
        np.testing.assert_allclose(result_from_source_model[key], result_from_passband_group[key])

    # Using toy passband group:
    toy_passband_group = create_toy_passband_group(tmp_path, delta_wave=20, trim_quantile=None)
    n_toy_bands = len(toy_passband_group.passbands)
    fluxes_source_model = model.get_band_fluxes(
        toy_passband_group,
        times=np.repeat(test_times, n_toy_bands),
        filters=np.tile(list(toy_passband_group.passbands.keys()), n_times),
        state=state,
    )
    result_from_source_model = {
        filter_name: fluxes_source_model[i::n_toy_bands]
        for i, filter_name in enumerate(toy_passband_group.passbands)
    }
    evaluated_fluxes = model.evaluate(test_times, toy_passband_group.waves, state)
    result_from_passband_group = toy_passband_group.fluxes_to_bandfluxes(evaluated_fluxes)

    # Check the two dicts are the same
    assert result_from_source_model.keys() == result_from_passband_group.keys()
    for key in result_from_source_model:
        np.testing.assert_allclose(result_from_source_model[key], result_from_passband_group[key])


def test_passband_group_calculate_in_band_wave_indices(passbands_dir, tmp_path):
    """Test the calculate_in_band_wave_indices method of the PassbandGroup class using both toy and LSST
    transmission tables."""
    # Using LSST passband group:
    lsst_passband_group = create_lsst_passband_group(passbands_dir, delta_wave=20, trim_quantile=None)

    # Make sure group.waves contains the union of all passband waves with no duplicates or other values
    np.testing.assert_allclose(
        lsst_passband_group.waves,
        np.unique(np.concatenate([passband.waves for passband in lsst_passband_group.passbands.values()])),
    )

    # Using toy passband group:
    toy_passband_group = create_toy_passband_group(tmp_path, delta_wave=20, trim_quantile=None)

    passband_A = toy_passband_group.passbands["TOY_a"]
    passband_B = toy_passband_group.passbands["TOY_b"]
    passband_C = toy_passband_group.passbands["TOY_c"]

    # Note that passband_A and passband_B have overlapping wavelength ranges
    # Where passband_A covers 100-300 and passband_B covers 250-350 (and passband_C covers 400-600)
    toy_a_inds = toy_passband_group._in_band_wave_indices["TOY_a"]
    np.testing.assert_allclose(toy_a_inds, np.array([0, 1, 2, 3, 4, 5, 6, 7, 9, 11, 13]))
    np.testing.assert_allclose(passband_A.waves, toy_passband_group.waves[toy_a_inds])

    toy_b_inds = toy_passband_group._in_band_wave_indices["TOY_b"]
    np.testing.assert_allclose(toy_b_inds, np.array([8, 10, 12, 14, 15, 16]))
    np.testing.assert_allclose(passband_B.waves, toy_passband_group.waves[toy_b_inds])

    toy_c_inds = toy_passband_group._in_band_wave_indices["TOY_c"]
    assert toy_c_inds == slice(17, 28)
    np.testing.assert_allclose(passband_C.waves, toy_passband_group.waves[toy_c_inds])



import numpy as np
from tdastro.astro_utils.mag_flux import flux2mag
from tdastro.astro_utils.zeropoint import flux_electron_zeropoint, magnitude_electron_zeropoint
from tdastro.opsim.opsim import LSSTCAM_PIXEL_SCALE, _lsstcam_dark_current, _lsstcam_readout_noise


def test_magnitude_electron_zeropoint():
    """Test that instrumental zeropoints are correct"""
    # Reproducing magnitude corresponding to S/N=5 from
    # https://smtn-002.lsst.io/v/OPSIM-1171/index.html
    fwhm_eff = {
        "u": 0.92,
        "g": 0.87,
        "r": 0.83,
        "i": 0.80,
        "z": 0.78,
        "y": 0.76,
    }
    fwhm_eff_getter = np.vectorize(fwhm_eff.get)
    sky_brightness = {
        "u": 23.05,
        "g": 22.25,
        "r": 21.20,
        "i": 20.46,
        "z": 19.61,
        "y": 18.60,
    }
    sky_brightness_getter = np.vectorize(sky_brightness.get)

    bands = list(fwhm_eff.keys())
    exptime = 30
    airmass = 1
    s2n = 5
    zp = magnitude_electron_zeropoint(
        band=bands, airmass=airmass, exptime=exptime, instr_zp=None, ext_coeff=None
    )
    sky_count_per_arcsec_sq = np.power(10.0, -0.4 * (sky_brightness_getter(bands) - zp))
    readout_per_arcsec_sq = _lsstcam_readout_noise**2 / LSSTCAM_PIXEL_SCALE**2
    dark_per_arcsec_sq = _lsstcam_dark_current * exptime / LSSTCAM_PIXEL_SCALE**2
    count_per_arcsec_sq = sky_count_per_arcsec_sq + readout_per_arcsec_sq + dark_per_arcsec_sq

    area = 2.266 * fwhm_eff_getter(bands) ** 2  # effective seeing area in arcsec^2
    n_background = count_per_arcsec_sq * area
    # sky-dominated regime would be n_signal = s2n * np.sqrt(n_background)
    n_signal = 0.5 * (s2n**2 + np.sqrt(s2n**4 + 4 * s2n**2 * n_background))
    mag_signal = zp - 2.5 * np.log10(n_signal)

    m5_desired = {
        "u": 23.70,
        "g": 24.97,
        "r": 24.52,
        "i": 24.13,
        "z": 23.56,
        "y": 22.55,
    }
    assert list(m5_desired) == bands
    m5_desired_getter = np.vectorize(m5_desired.get)

    np.testing.assert_allclose(mag_signal, m5_desired_getter(bands), atol=0.1)


def test_magnitude_electron_zeropoint_docstring():
    """Check if magnitude_electron_zeropoint has a docstring"""
    assert magnitude_electron_zeropoint.__doc__ is not None
    assert len(magnitude_electron_zeropoint.__doc__) > 100


def test_flux_electron_zeropoint():
    """Test that bandflux zeropoints are correct"""
    # Here we just check that magnitude-flux conversion is correct
    airmass = np.array([1, 1.5, 2]).reshape(-1, 1, 1)
    exptime = np.array([30, 38, 45]).reshape(1, -1, 1)
    bands = ["u", "g", "r", "i", "z", "y"]
    mag = magnitude_electron_zeropoint(
        band=bands, airmass=airmass, exptime=exptime, instr_zp=None, ext_coeff=None
    )
    flux = flux_electron_zeropoint(
        band=bands, airmass=airmass, exptime=exptime, instr_zp_mag=None, ext_coeff=None
    )
    np.testing.assert_allclose(mag, flux2mag(flux), rtol=1e-10)


def test_flux_electron_zeropoint_docstring():
    """Check if flux_electron_zeropoint has a docstring"""
    assert flux_electron_zeropoint.__doc__ is not None
    assert len(flux_electron_zeropoint.__doc__) > 100



import tempfile
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from tdastro.astro_utils.mag_flux import mag2flux
from tdastro.astro_utils.zeropoint import (
    _lsstcam_extinction_coeff,
    _lsstcam_zeropoint_per_sec_zenith,
)
from tdastro.opsim.opsim import (
    OpSim,
    create_random_opsim,
    opsim_add_random_data,
    oversample_opsim,
)


def test_create_opsim():
    """Create a minimal OpSim object and perform basic queries."""
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
    }
    pdf = pd.DataFrame(values)

    ops_data = OpSim(pdf)
    assert len(ops_data) == 5
    assert len(ops_data.columns) == 4

    # We have all the attributes set at their default values.
    assert ops_data.dark_current == 0.2
    assert ops_data.ext_coeff == _lsstcam_extinction_coeff
    assert ops_data.pixel_scale == 0.2
    assert ops_data.radius == 1.75
    assert ops_data.read_noise == 8.8
    assert ops_data.zp_per_sec == _lsstcam_zeropoint_per_sec_zenith

    # Check that we can extract the time bounds.
    t_min, t_max = ops_data.time_bounds()
    assert t_min == 0.0
    assert t_max == 4.0

    # We can query which columns the OpSim has.
    assert ops_data.has_columns("fieldRA")
    assert ops_data.has_columns("dec")
    assert not ops_data.has_columns("is_good_obs")
    assert ops_data.has_columns(["ra", "dec", "time"])
    assert not ops_data.has_columns(["ra", "dec", "vacation_time"])

    # We can access columns directly as though it was a table.
    assert np.allclose(ops_data["fieldRA"], values["fieldRA"])
    assert np.allclose(ops_data["fieldDec"], values["fieldDec"])
    assert np.allclose(ops_data["observationStartMJD"], values["observationStartMJD"])

    # Without a filters column we cannot access the filters.
    with pytest.raises(KeyError):
        _ = ops_data.get_filters()

    # We can create an OpSim directly from the dictionary as well.
    ops_data2 = OpSim(pdf)
    assert len(ops_data2) == 5
    assert len(ops_data.columns) == 4
    assert np.allclose(ops_data2["fieldRA"], values["fieldRA"])
    assert np.allclose(ops_data2["fieldDec"], values["fieldDec"])
    assert np.allclose(ops_data2["observationStartMJD"], values["observationStartMJD"])

    # We raise an error if we are missing a required row.
    del values["fieldDec"]
    with pytest.raises(KeyError):
        _ = OpSim(values)


def test_create_opsim_override():
    """Test that we can override the default survey values."""
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    ops_data = OpSim(
        values,
        dark_current=0.1,
        ext_coeff={"u": 0.1, "g": 0.2, "r": 0.3, "i": 0.4, "z": 0.5, "y": 0.6},
        pixel_scale=0.1,
        radius=1.0,
        read_noise=5.0,
        zp_per_sec={"u": 25.0, "g": 26.0, "r": 27.0, "i": 28.0, "z": 29.0, "y": 30.0},
    )

    # We have loaded the non-default values.
    assert ops_data.dark_current == 0.1
    assert ops_data.ext_coeff == {"u": 0.1, "g": 0.2, "r": 0.3, "i": 0.4, "z": 0.5, "y": 0.6}
    assert ops_data.pixel_scale == 0.1
    assert ops_data.radius == 1.0
    assert ops_data.read_noise == 5.0
    assert ops_data.zp_per_sec == {"u": 25.0, "g": 26.0, "r": 27.0, "i": 28.0, "z": 29.0, "y": 30.0}

    # We can access the filters.
    filters = ops_data.get_filters()
    assert set(filters) == {"r", "g", "i"}


def test_create_opsim_no_zp():
    """Create an opsim without a zeropoint column."""
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
    }

    # We fail if we do not have the other columns needed: filter, airmass, exptime.
    with pytest.raises(ValueError):
        _ = OpSim(values)

    values["filter"] = np.array(["r", "g", "r", "i", "z"])
    values["airmass"] = 0.01 * np.ones(5)
    values["visitExposureTime"] = 0.1 * np.ones(5)
    opsim = OpSim(values)

    assert opsim.has_columns("zp_nJy")
    assert np.all(opsim["zp_nJy"] >= 0.0)


def test_create_opsim_custom_names():
    """Create a minimal OpSim object from alternate column names."""
    values = {
        "custom_time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "custom_ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "custom_dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
    }

    # Load fails if we use the default colmap.
    with pytest.raises(KeyError):
        _ = OpSim(values)

    # Load succeeds if we pass in a customer dictionary.
    colmap = {"ra": "custom_ra", "dec": "custom_dec", "time": "custom_time"}
    ops_data = OpSim(values, colmap)
    assert len(ops_data) == 5


def test_opsim_add_columns():
    """Create a minimal OpSim object and perform basic queries."""
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
    }
    pdf = pd.DataFrame(values)

    ops_data = OpSim(pdf)
    assert len(ops_data) == 5
    assert len(ops_data.columns) == 4

    # We can add a column of constant values from a scalar.
    assert "new_column1" not in ops_data.columns
    ops_data.add_column("new_column1", 10)
    assert "new_column1" in ops_data.columns
    assert np.allclose(ops_data["new_column1"], [10, 10, 10, 10, 10])

    # We can add a column of data from a list or array.
    assert "new_column2" not in ops_data.columns
    ops_data.add_column("new_column2", [1, 2, 3, 4, 5])
    assert "new_column2" in ops_data.columns
    assert np.allclose(ops_data["new_column2"], [1, 2, 3, 4, 5])

    # We fail if we add a column with the incorrect number of values.
    with pytest.raises(ValueError):
        ops_data.add_column("new_column3", [1, 2, 3, 4, 5, 6])

    # We fail if we try to overwrite a current column unless we
    # set overwrite=True.
    with pytest.raises(KeyError):
        ops_data.add_column("new_column1", 12)
    assert np.allclose(ops_data["new_column1"], [10, 10, 10, 10, 10])

    ops_data.add_column("new_column2", 12, overwrite=True)
    assert np.allclose(ops_data["new_column2"], [12, 12, 12, 12, 12])

    # We can add random data.
    opsim_add_random_data(ops_data, "new_column3", min_val=1.0, max_val=5.0)
    values = ops_data["new_column3"]
    assert len(np.unique(values)) == 5
    assert np.all(values >= 1.0)
    assert np.all(values <= 5.0)


def test_opsim_filter_rows():
    """Test that the user can filter out OpSim rows."""
    times = np.arange(0.0, 10.0, 1.0)
    values = {
        "observationStartMJD": times,
        "fieldRA": 15.0 * (times + 1.0),
        "fieldDec": -1.0 * times,
        "zp_nJy": np.ones(10),
        "filter": np.tile(["r", "g"], 5),
    }
    ops_data = OpSim(values)
    assert len(ops_data) == 10
    assert len(ops_data.columns) == 5

    # We can filter the OpSim to specific rows by index.
    inds = [0, 1, 2, 3, 4, 5, 7, 8]
    ops_data = ops_data.filter_rows(inds)
    assert len(ops_data) == 8
    assert len(ops_data.columns) == 5

    expected_times = np.array(inds)
    assert np.allclose(ops_data["time"], expected_times)
    assert np.allclose(ops_data["ra"], 15.0 * (expected_times + 1.0))
    assert np.allclose(ops_data["dec"], -1.0 * expected_times)
    assert np.array_equal(ops_data["filter"], values["filter"][inds])

    # Check that the size of the internal KD-tree has changed.
    assert ops_data._kd_tree.n == 8

    # We can filter the OpSim to specific rows by mask.
    ops_data = ops_data.filter_rows(ops_data["filter"] == "r")
    assert len(ops_data) == 4

    expected_times = np.array([0.0, 2.0, 4.0, 8.0])
    assert np.allclose(ops_data["time"], expected_times)
    assert np.allclose(ops_data["ra"], 15.0 * (expected_times + 1.0))
    assert np.allclose(ops_data["dec"], -1.0 * expected_times)
    assert np.all(ops_data["filter"] == "r")

    # Check that the size of the internal KD-tree has changed (again).
    assert ops_data._kd_tree.n == 4

    # We throw an error if the mask is the wrong size.
    bad_mask = [True] * (len(ops_data) - 1)
    with pytest.raises(ValueError):
        _ = ops_data.filter_rows(bad_mask)


def test_read_small_opsim(opsim_small):
    """Read in a small OpSim file from the testing data directory."""
    ops_data = OpSim.from_db(opsim_small)
    assert len(ops_data) == 300


def test_write_read_opsim():
    """Create a minimal opsim data frame, test that we can write it,
    and test that we can correctly read it back in."""

    # Create a fake opsim data frame with just time, RA, and dec.
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
    }
    ops_data = OpSim(pd.DataFrame(values))

    with tempfile.TemporaryDirectory() as dir_name:
        file_path = Path(dir_name, "test_write_read_opsim.db")

        # The opsim does not exist until we write it.
        assert not file_path.is_file()
        with pytest.raises(FileNotFoundError):
            _ = OpSim.from_db(file_path)

        # We can write the opsim db.
        ops_data.write_opsim_table(file_path)
        assert file_path.is_file()

        # We can reread the opsim db.
        ops_data2 = OpSim.from_db(file_path)
        assert len(ops_data2) == 5
        assert np.allclose(values["observationStartMJD"], ops_data2["observationStartMJD"].to_numpy())
        assert np.allclose(values["fieldRA"], ops_data2["fieldRA"].to_numpy())
        assert np.allclose(values["fieldDec"], ops_data2["fieldDec"].to_numpy())

        # We cannot overwrite unless we set overwrite=True
        with pytest.raises(ValueError):
            ops_data.write_opsim_table(file_path, overwrite=False)
        ops_data.write_opsim_table(file_path, overwrite=True)


def test_opsim_range_search():
    """Test that we can extract the time, ra, and dec from an opsim data frame."""
    # Create a fake opsim data frame with just time, RA, and dec.
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]),
        "fieldRA": np.array([15.0, 15.0, 15.01, 15.0, 25.0, 24.99, 60.0, 5.0]),
        "fieldDec": np.array([-10.0, 10.0, 10.01, 9.99, 10.0, 9.99, -5.0, -1.0]),
        "zp_nJy": np.ones(8),
    }
    ops_data = OpSim(values)

    # Test single queries.
    assert set(ops_data.range_search(15.0, 10.0, 0.5)) == set([1, 2, 3])
    assert set(ops_data.range_search(25.0, 10.0, 0.5)) == set([4, 5])
    assert set(ops_data.range_search(15.0, 10.0, 100.0)) == set([0, 1, 2, 3, 4, 5, 6, 7])
    assert set(ops_data.range_search(15.0, 10.0, 1e-6)) == set([1])
    assert set(ops_data.range_search(15.02, 10.0, 1e-6)) == set()

    # With no radius provided, it should default to 1.75.
    assert set(ops_data.range_search(15.0, 10.0)) == set([1, 2, 3])
    assert set(ops_data.range_search(25.0, 10.0)) == set([4, 5])

    # Test is_observed() with single queries.
    assert ops_data.is_observed(15.0, 10.0, 0.5)
    assert not ops_data.is_observed(15.02, 10.0, 1e-6)

    # Test a batched query.
    query_ra = np.array([15.0, 25.0, 15.0])
    query_dec = np.array([10.0, 10.0, 5.0])
    neighbors = ops_data.range_search(query_ra, query_dec, 0.5)
    assert len(neighbors) == 3
    assert set(neighbors[0]) == set([1, 2, 3])
    assert set(neighbors[1]) == set([4, 5])
    assert set(neighbors[2]) == set()

    # Test is_observed() with batched queries.
    assert np.array_equal(
        ops_data.is_observed(query_ra, query_dec, 0.5),
        np.array([True, True, False]),
    )

    # Test that we fail if bad query arrays are provided.
    with pytest.raises(ValueError):
        _ = ops_data.range_search(None, None, 0.5)
    with pytest.raises(ValueError):
        _ = ops_data.range_search([1.0, 2.3], 4.5, 0.5)
    with pytest.raises(ValueError):
        _ = ops_data.range_search([1.0, 2.3], [4.5, 6.7, 8.9], 0.5)
    with pytest.raises(ValueError):
        _ = ops_data.range_search([1.0, 2.3], [4.5, None], 0.5)


def test_opsim_get_observations():
    """Test that we can extract the time, ra, and dec from an opsim data frame."""
    # Create a fake opsim data frame with just time, RA, and dec.
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]),
        "fieldRA": np.array([15.0, 15.0, 15.01, 15.0, 25.0, 24.99, 60.0, 5.0]),
        "fieldDec": np.array([-10.0, 10.0, 10.01, 9.99, 10.0, 9.99, -5.0, -1.0]),
        "zp_nJy": np.ones(8),
    }
    ops_data = OpSim(values)

    # Test basic queries (all columns).
    obs = ops_data.get_observations(15.0, 10.0, 0.5)
    assert len(obs) == 4
    assert np.allclose(obs["observationStartMJD"], [1.0, 2.0, 3.0])

    obs = ops_data.get_observations(25.0, 10.0, 0.5)
    assert np.allclose(obs["observationStartMJD"], [4.0, 5.0])

    obs = ops_data.get_observations(15.0, 10.0, 100.0)
    assert np.allclose(obs["observationStartMJD"], [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0])

    obs = ops_data.get_observations(15.0, 10.0, 1e-6)
    assert np.allclose(obs["observationStartMJD"], [1.0])

    obs = ops_data.get_observations(15.02, 10.0, 1e-6)
    assert len(obs["observationStartMJD"]) == 0

    # Test we can get a subset of columns.
    obs = ops_data.get_observations(15.0, 10.0, 0.5, cols=["observationStartMJD", "zp_nJy"])
    assert len(obs) == 2
    assert np.allclose(obs["observationStartMJD"], [1.0, 2.0, 3.0])

    # Test we can use the colmap names.
    obs = ops_data.get_observations(15.0, 10.0, 0.5, cols=["time", "ra"])
    assert len(obs) == 2
    assert np.allclose(obs["time"], [1.0, 2.0, 3.0])

    # Test we fail with an unrecognized column name.
    with pytest.raises(KeyError):
        _ = ops_data.get_observations(15.0, 10.0, 0.5, cols=["time", "custom_col"])


def test_opsim_docstring():
    """Test if OpSim class has a docstring"""
    assert OpSim.__doc__ is not None
    assert len(OpSim.__doc__) > 100


def test_read_opsim_shorten(opsim_shorten):
    """Read in a shorten OpSim file from the testing data directory."""
    ops_data = OpSim.from_db(opsim_shorten)
    assert len(ops_data) == 100


def test_opsim_flux_err_point_source(opsim_shorten):
    """Check if OpSim.flux_err_point_source is consistent with fiveSigmaDepth."""
    ops_data = OpSim.from_db(opsim_shorten)
    # fiveSigmaDepth is the 5-sigma limiting magnitude.
    flux = mag2flux(ops_data.table["fiveSigmaDepth"])
    expected_flux_err = flux / 5.0

    flux_err = ops_data.bandflux_error_point_source(flux, index=np.arange(len(ops_data)))

    # Tolerance is very high, we should investigate why the values are so different.
    np.testing.assert_allclose(flux_err, expected_flux_err, rtol=0.2)


def test_create_random_opsim():
    """Test that we can create a complete random OpSim."""
    opsim = create_random_opsim(1000)
    assert len(opsim) == 1000


def test_oversample_opsim(opsim_shorten):
    """Test that we can oversample an OpSim file."""
    opsim = OpSim.from_db(opsim_shorten)

    bands = ["g", "r"]
    ra, dec = 205.0, -57.0
    time_range = 60_000.0, 60_010.0
    delta_t = 0.01

    for strategy in ["darkest_sky", "random"]:
        oversampled = oversample_opsim(
            opsim,
            pointing=(ra, dec),
            time_range=time_range,
            delta_t=delta_t,
            bands=bands,
            strategy=strategy,
        )
        assert set(opsim.table.columns) == set(oversampled.table.columns), "columns are not the same"
        np.testing.assert_allclose(
            np.diff(oversampled["observationStartMJD"]), delta_t, err_msg="delta_t is not correct"
        )
        np.testing.assert_allclose(oversampled["fieldRA"], ra, err_msg="RA is not correct")
        np.testing.assert_allclose(oversampled["fieldDec"], dec, err_msg="Dec is not correct")
        assert np.all(oversampled["observationStartMJD"] >= time_range[0]), "time range is not correct"
        assert np.all(oversampled["observationStartMJD"] <= time_range[1]), "time range is not correct"
        assert set(oversampled["filter"]) == set(bands), "oversampled table has the wrong bands"
        assert (
            oversampled["skyBrightness"].unique().size >= oversampled["filter"].unique().size
        ), "there should be at least as many skyBrightness values as bands"
        assert oversampled["skyBrightness"].isna().sum() == 0, "skyBrightness has NaN values"

    # Oversampling fails if there are no observations in the time range.
    with pytest.raises(ValueError):
        _ = oversample_opsim(
            opsim,
            pointing=(5.0, 67.0),
            time_range=(0.0, 1.0),
            delta_t=delta_t,
            bands=bands,
            strategy=strategy,
        )

    # Oversampling fails with an invalid strategy.
    with pytest.raises(ValueError):
        _ = oversample_opsim(
            opsim,
            pointing=(ra, dec),
            time_range=(0.0, 1.0),
            delta_t=delta_t,
            bands=bands,
            strategy="invalid",
        )


def test_fixture_oversampled_observations(oversampled_observations):
    """Test the fixture oversampled_observations."""
    assert len(oversampled_observations) == 36_500
    assert set(oversampled_observations["filter"]) == {"g", "r"}
    assert oversampled_observations["skyBrightness"].isna().sum() == 0
    assert oversampled_observations["skyBrightness"].unique().size >= 2
    assert np.all(oversampled_observations["observationStartMJD"] >= 61406.0)
    assert np.all(oversampled_observations["observationStartMJD"] <= 61771.0)
    np.testing.assert_allclose(oversampled_observations["fieldRA"], 0.0)
    np.testing.assert_allclose(oversampled_observations["fieldDec"], 0.0)
    np.testing.assert_allclose(np.diff(oversampled_observations["observationStartMJD"]), 0.01)



import numpy as np
import pytest
from tdastro.sources.periodic_source import PeriodicSource


class SineSource(PeriodicSource):
    """A simple sine source with power (Rayleigh–Jeans) spectrum."""

    def _evaluate_phases(self, phases, wavelengths, graph_state, **kwargs):
        del kwargs

        amplitude = 2 + np.sin(2 * np.pi * phases[:, None])
        return amplitude * wavelengths[None, :] ** -2


@pytest.mark.parametrize("period, t0", [(1.0, 0.0), (2.0, 0.75), (4.0, 100 / 3)])
def test_periodicity(period, t0):
    """Test that the source is periodic."""
    max_time = 16
    n_periods = int(max_time / period)

    source = SineSource(period=period, t0=t0)
    times = np.linspace(0, max_time, max_time * 100 + 1)
    wavelengths = np.linspace(100, 200, 3)
    fluxes = source.evaluate(times, wavelengths)

    # Test that the flux is periodic on boundaries.
    np.testing.assert_allclose(fluxes[0], fluxes[-1])

    # Test that each period is identical.
    # First :-1 is to drop the last point that belongs to the next period.
    fluxes_in_periods = fluxes[:-1].reshape(n_periods, -1, len(wavelengths))
    for i_period in range(1, n_periods):
        np.testing.assert_allclose(fluxes_in_periods[0], fluxes_in_periods[i_period])



import matplotlib.pyplot as plt
import numpy as np
import pytest
from tdastro.utils.plotting import plot_bandflux_lightcurves, plot_flux_spectrogram, plot_lightcurves


def test_plot_lightcurves():
    """Test that we can plot light curves."""
    # Test minimal input
    fluxes = np.array([1.0, 2.0, 3.0])
    times = np.array([1.0, 2.0, 3.0])
    plot_lightcurves(fluxes, times)

    # ValueError if len(times) != len(fluxes)
    wrong_times = np.array([1.0, 2.0])
    with pytest.raises(ValueError):
        plot_lightcurves(fluxes, wrong_times)

    # ValueError if len(filters) != len(fluxes)
    wrong_filters = ["none"]
    with pytest.raises(ValueError):
        plot_lightcurves(fluxes, times, filters=wrong_filters)

    # ValueError if fluxerrs given and len(fluxerrs) != len(fluxes)
    wrong_fluxerrs = np.array([0.1, 0.2])
    with pytest.raises(ValueError):
        plot_lightcurves(fluxes, times, fluxerrs=wrong_fluxerrs)

    # Test with almost all inputs given:
    # - fluxerrs (same length as fluxes to pass the ValueError check)
    # - filters (same length as fluxes to pass the ValueError check)
    # - title
    fluxerrs = np.array([0.1, 0.2, 0.3])
    filters = np.array(["A", "B", "A"])
    title = "Test Title"
    plot_lightcurves(fluxes, times, fluxerrs=fluxerrs, filters=filters, title=title)

    # Test with all inputs given:
    # - ax (matplotlib axes object)
    # - figure (matplotlib figure object)
    fig, ax = plt.subplots()
    plot_lightcurves(fluxes, times, fluxerrs=fluxerrs, filters=filters, ax=ax, figure=fig, title=title)
    plt.close(fig)


def test_plot_bandflux_lightcurves():
    """Test that we can plot bandflux light curves."""
    # Test minimal input
    bandfluxes = np.array([[0.0, 1.0, 2.0, 3.0], [4.0, 5.0, 6.0, 7.0]])
    plot_bandflux_lightcurves(bandfluxes)

    # Test with all inputs given:
    # - times (np array of length T)
    # - ax (matplotlib axes object)
    # - figure (matplotlib figure object)
    # - title
    times = np.array([1.0, 2.0])
    fig, ax = plt.subplots()
    title = "Test Title"
    plot_bandflux_lightcurves(bandfluxes, times=times, ax=ax, figure=fig, title=title)
    plt.close(fig)


def test_plot_flux_spectrogram():
    """Test that we can plot a flux spectrogram."""
    # Test minimal input
    fluxes = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    plot_flux_spectrogram(fluxes)

    # Test with all inputs given:
    # - times (np array)
    # - ax (matplotlib axes object)
    # - figure (matplotlib figure object)
    # - title
    times = np.array([1.0, 2.0, 3.0, 4.0])
    fig, ax = plt.subplots()
    title = "Test Title"
    plot_flux_spectrogram(fluxes, times=times, ax=ax, figure=fig, title=title)
    plt.close(fig)



import random

import jax
import numpy as np
import pytest
from tdastro.base_models import FunctionNode, ParameterizedNode, ParameterSource
from tdastro.math_nodes.single_value_node import SingleVariableNode


def _sampler_fun(**kwargs):
    """Return a random value between 0 and 1.0.

    Parameters
    ----------
    **kwargs : `dict`, optional
        Absorbs additional parameters
    """
    return random.random()


def _test_func(value1, value2):
    """Return the sum of the two parameters.

    Parameters
    ----------
    value1 : `float`
        The first parameter.
    value2 : `float`
        The second parameter.
    """
    return value1 + value2


class PairModel(ParameterizedNode):
    """A test class for the ParameterizedNode.

    Parameters
    ----------
    value1 : `float`
        The first value.
    value2 : `float`
        The second value.
    value_sum : `float`
        The sum of the two values.
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    def __init__(self, value1, value2, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter("value1", value1, **kwargs)
        self.add_parameter("value2", value2, **kwargs)
        self.add_parameter(
            "value_sum",
            FunctionNode(_test_func, value1=self.value1, value2=self.value2),
            **kwargs,
        )


def test_parameter_source():
    """Test the ParameterSource creation and setter functions."""
    source = ParameterSource("test")
    assert source.parameter_name == "test"
    assert source.node_name == ""
    assert source.source_type == ParameterSource.UNDEFINED
    assert source.dependency is None
    assert source.value is None

    source.set_as_constant(10.0)
    assert source.parameter_name == "test"
    assert source.source_type == ParameterSource.CONSTANT
    assert source.dependency is None
    assert source.value == 10.0

    with pytest.raises(ValueError):
        source.set_as_constant(_test_func)

    node = SingleVariableNode("A", 20.0)
    source.set_as_parameter(node, "A")
    assert source.source_type == ParameterSource.MODEL_PARAMETER
    assert source.dependency is node
    assert source.value == "A"

    func = FunctionNode(_test_func, value1=0.1, value2=0.4)
    source.set_as_function(func)
    assert source.source_type == ParameterSource.FUNCTION_NODE
    assert source.dependency is func


def test_parameterized_node():
    """Test that we can sample and create a PairModel object."""
    # Simple addition
    model1 = PairModel(value1=0.5, value2=0.5)
    assert str(model1) == "PairModel"

    # Check that we can lookup parameters.
    assert model1.has_valid_param("value1")
    assert model1.has_valid_param("value2")
    assert model1.has_valid_param("value_sum")
    assert not model1.has_valid_param("value3")

    state = model1.sample_parameters()
    assert model1.get_param(state, "value1") == 0.5
    assert model1.get_param(state, "value2") == 0.5
    assert model1.get_param(state, "value_sum") == 1.0

    # We can also access the parameters using their names and the graph state.
    assert model1.value1(state) == 0.5
    assert model1.value2(state) == 0.5
    assert model1.value_sum(state) == 1.0

    # We use the default for un-assigned parameters.
    assert model1.get_param(state, "value3") is None
    assert model1.get_param(state, "value4", 1.0) == 1.0

    # Check that we can list the parameters.
    assert np.array_equal(model1.list_params(), ["value1", "value2", "value_sum"])

    # If we set a position (and there is no node_label), the position shows up in the name.
    model1.node_pos = 100
    model1._update_node_string()
    assert str(model1) == "PairModel_100"

    # Use value1=model.value and value2=1.0
    model2 = PairModel(value1=model1.value1, value2=1.0, node_label="test")
    assert str(model2) == "test"

    state = model2.sample_parameters()
    assert model2.get_param(state, "value1") == 0.5
    assert model2.get_param(state, "value2") == 1.0
    assert model2.get_param(state, "value_sum") == 1.5

    # Compute value1 from model2's result and value2 from the sampler function.
    # The sampler function is auto-wrapped in a FunctionNode.
    model3 = PairModel(value1=model2.value_sum, value2=_sampler_fun)
    state = model3.sample_parameters()
    rand_val = model3.get_param(state, "value2")
    assert model3.get_param(state, "value_sum") == pytest.approx(1.5 + rand_val)

    # Compute value1 from model3's result (which is itself the result for model2 +
    # a random value) and value2 = -1.0.
    model4 = PairModel(value1=model3.value_sum, value2=-1.0)
    state = model4.sample_parameters()
    rand_val = model3.get_param(state, "value2")
    assert model4.get_param(state, "value_sum") == pytest.approx(0.5 + rand_val)

    # We can resample and it should change the result.
    new_state = model4.sample_parameters()
    rand_val = model3.get_param(new_state, "value2")

    # Nothing changes in model1 or model2
    assert model1.get_param(new_state, "value1") == 0.5
    assert model1.get_param(new_state, "value2") == 0.5
    assert model1.get_param(new_state, "value_sum") == 1.0
    assert model2.get_param(new_state, "value1") == 0.5
    assert model2.get_param(new_state, "value2") == 1.0
    assert model2.get_param(new_state, "value_sum") == 1.5

    # Models 3 and 4 use the data from the new random value.
    assert model3.get_param(new_state, "value_sum") == pytest.approx(1.5 + rand_val)
    assert model4.get_param(new_state, "value_sum") == pytest.approx(0.5 + rand_val)
    assert model4.get_param(state, "value_sum") != model4.get_param(new_state, "value_sum")

    # Test that we give the default if the parameter is not found.
    assert model1.get_param(state, "no_such_param") is None
    assert model1.get_param(state, "no_such_param", 2.0) == 2.0

    # Check that we fail if we give an invalid GraphState.
    with pytest.raises(ValueError):
        _ = model1.get_param(None, "value_sum")
    with pytest.raises(ValueError):
        _ = model1.get_local_params(None)


def test_parameterized_node_label_collision():
    """Test that throw an error when two nodes use the same label."""
    node_a = SingleVariableNode("A", 10.0, node_label="A")
    node_b = SingleVariableNode("B", 20.0, node_label="B")
    pair1 = PairModel(value1=node_a.A, value2=node_b.B, node_label="pair1")
    pair2 = PairModel(value1=pair1.value_sum, value2=node_b.B, node_label="pair2")

    # No collision even though node_b is referenced twice.
    state = pair2.sample_parameters()
    assert state["pair1"]["value_sum"] == 30.0
    assert state["pair2"]["value_sum"] == 50.0

    # We run into a problem if we reuse a label. Label "A" collides multiple
    # levels above.
    node_c = SingleVariableNode("C", 5.0, node_label="C")
    pair3 = PairModel(value1=pair2.value_sum, value2=node_c.C, node_label="A")
    with pytest.raises(ValueError):
        _ = pair3.sample_parameters()


def test_parameterized_node_modify():
    """Test that we can modify the parameters in a node."""
    model = PairModel(value1=0.5, value2=0.5)

    # We cannot add a parameter a second time.
    with pytest.raises(KeyError):
        model.add_parameter("value1", 5.0)

    # We can set the parameter.
    model.set_parameter("value1", 5.0)

    # We cannot set a value that hasn't been added.
    with pytest.raises(KeyError):
        model.set_parameter("brightness", 5.0)


def test_parameterized_node_self_parameter():
    """Test that we fail if we try to use the same node as a dependency."""

    class SelfDepNode(ParameterizedNode):
        """A class with a loop in its dependencies."""

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.add_parameter("value1", self, **kwargs)

    with pytest.raises(ValueError):
        _ = SelfDepNode()


def test_parameterized_node_from_node():
    """Test that we can set the values of a parameterized node directly
    from the values of another parameterized node.
    """
    model1 = PairModel(value1=0.5, value2=1.5, node_label="A")
    model2 = PairModel(value1=model1, value2=model1, node_label="B")

    # The "value1" and "value2" of model2 should be the same as model1.
    state = model2.sample_parameters()
    assert state["B"]["value1"] == 0.5
    assert state["B"]["value2"] == 1.5

    # We fail if we try to reference a parameter in a node that does not have it.
    model3 = SingleVariableNode("value1", 0.5, node_label="C")
    with pytest.raises(ValueError):
        _ = PairModel(value1=model3, value2=model3)


def test_parameterized_node_overwrite_fun():
    """Test that we fail if we try to set a parameter that conflicts with
    something predefined in the class.  This prevents us overwriting functions
    by accident.
    """

    class OverwriteFunNode(ParameterizedNode):
        """A class with a function that conflicts with a parameter."""

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.add_parameter("sum", 1.0, **kwargs)

        def sum(self):
            """Return the sum."""
            return 2.0

    with pytest.raises(KeyError):
        _ = OverwriteFunNode()


def test_parameterized_node_build_pytree():
    """Test that we can extract the PyTree of a graph."""
    model1 = PairModel(value1=0.5, value2=1.5, node_label="A")
    model2 = PairModel(value1=model1.value1, value2=3.0, node_label="B")
    graph_state = model2.sample_parameters()

    pytree = model2.build_pytree(graph_state)
    assert pytree["A"]["value1"] == 0.5
    assert pytree["A"]["value2"] == 1.5
    assert pytree["B"]["value2"] == 3.0

    # Manually set value2 to allow_gradient to False and check that it no
    # longer appears in the pytree.
    model1.setters["value2"].allow_gradient = False

    pytree = model2.build_pytree(graph_state)
    assert pytree["A"]["value1"] == 0.5
    assert pytree["B"]["value2"] == 3.0
    assert "value2" not in pytree["A"]

    # If we set node B's value1 to allow the gradient, it will appear and
    # neither of node A's value will appear (because the gradient stops at
    # B.value1).
    model1.setters["value2"].allow_gradient = True
    model2.setters["value1"].allow_gradient = True

    pytree = model2.build_pytree(graph_state)
    assert "A" not in pytree
    assert pytree["B"]["value1"] == 0.5
    assert pytree["B"]["value2"] == 3.0


def test_function_node_basic():
    """Test that we can create and query a FunctionNode."""
    my_func = FunctionNode(_test_func, value1=1.0, value2=2.0)
    state = my_func.sample_parameters()

    assert my_func.compute(state) == 3.0
    assert my_func.compute(state, value2=3.0) == 4.0
    assert my_func.compute(state, value2=3.0, unused_param=5.0) == 4.0
    assert my_func.compute(state, value2=3.0, value1=1.0) == 4.0
    assert str(my_func) == "FunctionNode:_test_func_0"

    # We can also compute this result (for testing) by calling generate().
    assert my_func.generate() == 3.0


def test_function_node_chain():
    """Test that we can create and query a chained FunctionNode."""
    func1 = FunctionNode(_test_func, value1=1.0, value2=1.0)
    func2 = FunctionNode(_test_func, value1=func1, value2=3.0)
    state = func2.sample_parameters()

    assert func2.compute(state) == 5.0


def test_no_resample_functions():
    """Test that if we use the same node as dependencies in two other nodes, we do not resample it."""
    rand_val = FunctionNode(_sampler_fun)
    node_a = SingleVariableNode("A", rand_val)
    node_b = PairModel(value1=node_a.A, value2=rand_val)
    state = node_b.sample_parameters()

    # All the values should be the same (no resampling of A.)
    assert node_b.get_param(state, "value1") == node_a.get_param(state, "A")
    assert node_b.get_param(state, "value2") == node_a.get_param(state, "A")

    # They should change when we resample.
    new_state = node_b.sample_parameters()
    assert node_a.get_param(state, "A") != node_a.get_param(new_state, "A")
    assert node_b.get_param(new_state, "value1") == node_a.get_param(new_state, "A")
    assert node_b.get_param(new_state, "value2") == node_a.get_param(new_state, "A")


def test_np_sampler_method():
    """Test that we can wrap numpy random functions."""
    rng = np.random.default_rng(1001)
    my_func = FunctionNode(rng.normal, loc=10.0, scale=1.0, outputs=["val"])

    # Sample 1000 times with the default values. Check that we are near
    # the expected mean and not everything is equal.
    vals = []
    for _ in range(1000):
        state = my_func.sample_parameters()
        vals.append(my_func.get_param(state, "val"))
    assert abs(np.mean(vals) - 10.0) < 1.0
    assert not np.all(vals == vals[0])


def test_function_node_multi_dim():
    """Test that we can query a FunctionNode with multi-dimensional input."""
    my_func = FunctionNode(np.sum, a=[1, 2, 3, 4], axis=-1, node_label="sum")
    state = my_func.sample_parameters()
    assert state["sum"]["function_node_result"] == 10

    # Test with multiple samples. To make the sum meaningful we need to set which
    # axis we are summing across. We can do this with the fixed_params attribute.
    my_func = FunctionNode(np.sum, a=[1, 2, 3, 4, 5], fixed_params={"axis": 1}, node_label="sum")
    state = my_func.sample_parameters(num_samples=10)
    assert len(state["sum"]["function_node_result"]) == 10
    assert np.all(state["sum"]["function_node_result"] == 15)


def test_function_node_obj():
    """Test that we can create and query a FunctionNode that depends on
    another ParameterizedNode.
    """
    # The model depends on the function.
    func = FunctionNode(_test_func, value1=5.0, value2=6.0)
    model = PairModel(value1=10.0, value2=func)
    state = model.sample_parameters()
    assert model.get_param(state, "value_sum") == 21.0

    # Function depends on the model's value2 parameter.
    model = PairModel(value1=-10.0, value2=17.5)
    func = FunctionNode(_test_func, value1=5.0, value2=model.value2, outputs=["res"])
    state = func.sample_parameters()

    assert model.get_param(state, "value_sum") == 7.5
    assert func.get_param(state, "res") == 22.5

    # We can always override the node's parameters with kwargs.
    assert func.compute(state, value1=1.0, value2=4.0) == 5.0


def test_function_node_multi():
    """Test that we can query a function node with multiple outputs."""

    def _test_func2(value1, value2):
        return (value1 + value2, value1 - value2)

    my_func = FunctionNode(_test_func2, outputs=["sum_res", "diff_res"], value1=5.0, value2=6.0)
    model = PairModel(value1=my_func.sum_res, value2=my_func.diff_res)
    state = model.sample_parameters()

    assert model.get_param(state, "value1") == 11.0
    assert model.get_param(state, "value2") == -1.0
    assert model.get_param(state, "value_sum") == 10.0


def test_function_node_jax():
    """Test that we can perform a JAX grad computation on a graph of function nodes."""

    def _test_func2(value1, value2):
        return value1 / value2

    # Create a function of a / b + c
    div_node = FunctionNode(_test_func2, value1=4.0, value2=0.5, node_label="div")
    sum_node = FunctionNode(_test_func, value1=1.0, value2=div_node, node_label="sum")
    graph_state = sum_node.sample_parameters()

    pytree = sum_node.build_pytree(graph_state)
    gr_func = jax.value_and_grad(sum_node.generate)
    values, gradients = gr_func(pytree)
    assert values == 9.0
    assert gradients["sum"]["value1"] == 1.0
    assert gradients["div"]["value1"] == 2.0
    assert gradients["div"]["value2"] == -16.0



"""The ColorLaw used by SALT models as defined by in (Guy J., 2007)

It is adapted from sncosmo's SALT2ColorLaw class (but implemented in JAX):
https://github.com/sncosmo/sncosmo/blob/v2.10.1/sncosmo/salt2utils.pyx
"""

import jax.numpy as jnp
import numpy as np
from citation_compass import CiteClass

# Constants used in SALT2ColorLaw computations (from:
# https://github.com/sncosmo/sncosmo/blob/v2.10.1/sncosmo/salt2utils.pyx)
_SALT2CL_B = 4302.57  # B-band-ish wavelength
_SALT2CL_V = 5428.55  # V-band-ish wavelength
_WAVESCALE = 1.0 / (_SALT2CL_V - _SALT2CL_B)


class SALT2ColorLaw(CiteClass):
    """An object that applies the color law to the given wavelengths.

    References
    ----------
    sncosmo - https://zenodo.org/records/14714968

    Parameters
    ----------
    wave_min : float
        The minimum wavelength (in angstroms)
    wave_max : float
        The maximum wavelength (in angstroms)
    coeffs : list, numpy array, or jax array
        The <= 6 coefficients of the polynomial to use.

    Attributes
    ----------
    coeffs : JAX array
        The final 7 coefficients to use (based on the <= 6 provided in the parameters).
    scaled_wave_min : float
        The minimum wavelength shifted and scaled.
    scaled_wave_max : float
        The maximum wavelength shifted and scaled.
    value_at_min : float
        The value of the polynomial at wave_min.
    value_at_max : float
        The value of the polynomial at wave_max.
    _exponents : JAX array
        A precomputed array of the exponents to use.
    """

    def __init__(self, wave_min, wave_max, coeffs):
        # Create the internal coefficient array. The new first entry is 1.0 minus the
        # sum of the given entries. The first six given entries are then listed.
        coeffs = np.array(coeffs)
        num_coeffs = min(len(coeffs), 6)

        padded_coeffs = np.zeros(7)
        padded_coeffs[1 : (num_coeffs + 1)] = np.array(coeffs)[0:num_coeffs]
        padded_coeffs[0] = 1.0 - np.sum(padded_coeffs)
        self.coeffs = jnp.asarray(padded_coeffs)

        # Compute the bounds for the wavelengths and the value of the polynomial at both bounds.
        self.scaled_wave_min = (wave_min - _SALT2CL_B) * _WAVESCALE
        self.scaled_wave_max = (wave_max - _SALT2CL_B) * _WAVESCALE

        self.exponents = jnp.arange(1, 8, 1)
        self.value_at_min = jnp.sum(self.coeffs * jnp.power(self.scaled_wave_min, self.exponents))
        self.value_at_max = jnp.sum(self.coeffs * jnp.power(self.scaled_wave_max, self.exponents))

        # Precompute the polynomials derivative at the min and max wavelength.
        dcoeffs = jnp.arange(2, 8, 1) * self.coeffs[1:7]
        dexponents = jnp.arange(1, 7, 1)
        self.deriv_at_min = jnp.sum(dcoeffs * jnp.power(self.scaled_wave_min, dexponents)) + self.coeffs[0]
        self.deriv_at_max = jnp.sum(dcoeffs * jnp.power(self.scaled_wave_max, dexponents)) + self.coeffs[0]

    @classmethod
    def from_file(cls, filename):
        """Create the SALT2ColorLaw object from data in a file.

        Parameters
        ----------
        filename : str
            The name of the file to load.
        """
        with open(filename, mode="r") as f:
            data = f.read().split()

        # The first line holds the number of coefficients N and the next N lines
        # each hold a single coefficient.
        num_coeff = int(data[0])
        coeffs = np.array(data[1 : (1 + num_coeff)], dtype=float)

        # The rest of the lines (if any) are meta-data with a label and value on each line.
        wave_min = 3000.0
        wave_max = 7000.0
        for i in range(1 + num_coeff, len(data), 2):
            if "min_lambda" in data[i]:
                wave_min = float(data[i + 1])
            elif "max_lambda" in data[i]:
                wave_max = float(data[i + 1])
            elif "version" in data[i]:
                version = int(data[i + 1])
                if version != 1:
                    raise RuntimeError(f"Unsupported version {version}.")

        return SALT2ColorLaw(wave_min, wave_max, coeffs)

    def apply(self, wavelengths):
        """Apply the color law to the given wavelengths.

        Parameters
        ----------
        wavelengths : array
            The wavelengths in angstroms.
        """
        num_waves = len(wavelengths)
        shifted_wave = (jnp.asarray(wavelengths) - _SALT2CL_B) * _WAVESCALE

        # Compute the three cases of interest.
        # 1) If the shifted wave is past the lower bound, extrapolate a value based on
        #    the value and derivative at that bound.
        below = self.value_at_min + self.deriv_at_min * (shifted_wave - self.scaled_wave_min)
        # 2) If the shifted wave is past the upper bound, extrapolate a value based on
        #    the value and derivative at that bound.
        above = self.value_at_max + self.deriv_at_max * (shifted_wave - self.scaled_wave_max)
        # 3) If the shifted value is in the middle, use the polynomial.
        wave_T = jnp.reshape(shifted_wave, (num_waves, 1))
        coeffs_all = jnp.tile(self.coeffs, (num_waves, 1))
        middle = jnp.sum((coeffs_all * jnp.power(wave_T, self.exponents)).T, axis=0)

        result = -jnp.where(
            shifted_wave < self.scaled_wave_min,
            below,
            jnp.where(
                shifted_wave > self.scaled_wave_max,
                above,
                middle,
            ),
        )
        return result



import numpy as np
import pandas as pd
import pytest
from astropy.table import Table
from tdastro.base_models import FunctionNode
from tdastro.graph_state import GraphState
from tdastro.math_nodes.given_sampler import (
    GivenValueList,
    GivenValueSampler,
    GivenValueSelector,
    TableSampler,
)


def _test_func(value1, value2):
    """Return the sum of the two parameters.

    Parameters
    ----------
    value1 : `float`
        The first parameter.
    value2 : `float`
        The second parameter.
    """
    return value1 + value2


def test_given_value_list():
    """Test that we can retrieve numbers from a GivenValueList."""
    given_node = GivenValueList([1.0, 1.5, 2.0, 2.5, 3.0, -1.0, 3.5])

    # Check that we generate the correct result and save it in the GraphState.
    state1 = GraphState(num_samples=2)
    results = given_node.compute(state1)
    assert np.array_equal(results, [1.0, 1.5])
    assert np.array_equal(given_node.get_param(state1, "function_node_result"), [1.0, 1.5])

    state2 = GraphState(num_samples=1)
    results = given_node.compute(state2)
    assert results == 2.0
    assert given_node.get_param(state2, "function_node_result") == 2.0

    state3 = GraphState(num_samples=2)
    results = given_node.compute(state3)
    assert np.array_equal(results, [2.5, 3.0])
    assert np.array_equal(given_node.get_param(state3, "function_node_result"), [2.5, 3.0])

    # Check that GivenValueList raises an error when it has run out of samples.
    state4 = GraphState(num_samples=4)
    with pytest.raises(IndexError):
        _ = given_node.compute(state4)

    # Resetting the GivenValueList starts back at the beginning.
    given_node.reset()
    state5 = GraphState(num_samples=6)
    results = given_node.compute(state5)
    assert np.array_equal(results, [1.0, 1.5, 2.0, 2.5, 3.0, -1.0])
    assert np.array_equal(
        given_node.get_param(state5, "function_node_result"),
        [1.0, 1.5, 2.0, 2.5, 3.0, -1.0],
    )


def test_test_given_value_list_compound():
    """Test that we can use the GivenValueList as input into another node."""
    values = [1.0, 1.5, 2.0, 2.5, 3.0, -1.0, 3.5, 4.0, 10.0, -2.0]
    given_node = GivenValueList(values)

    # Create a function node that takes the next value and adds 2.0.
    compound_node = FunctionNode(_test_func, value1=given_node, value2=2.0)
    for val in values:
        state = compound_node.sample_parameters()
        assert compound_node.get_param(state, "function_node_result") == val + 2.0

    # Reset the given node and try generating all the samples.
    given_node.reset()
    state2 = compound_node.sample_parameters(num_samples=8)
    assert np.array_equal(
        compound_node.get_param(state2, "function_node_result"),
        [3.0, 3.5, 4.0, 4.5, 5.0, 1.0, 5.5, 6.0],
    )


def test_given_value_list_of_list():
    """Test that we can retrieve lists from a GivenValueList."""
    given_node = GivenValueList([[0, 1], [1, 2], [2, 3], [3, 4], [4, 5]], node_label="node")

    # Check that we generate the correct result and save it in the GraphState.
    state = given_node.sample_parameters(num_samples=3)
    assert len(state["node"]["function_node_result"]) == 3
    assert np.array_equal(state["node"]["function_node_result"][0], [0, 1])
    assert np.array_equal(state["node"]["function_node_result"][1], [1, 2])
    assert np.array_equal(state["node"]["function_node_result"][2], [2, 3])

    # When we generate a single sample, we should get a list instead of a list of lists.
    state2 = given_node.sample_parameters(num_samples=1)
    assert len(state2["node"]["function_node_result"]) == 2
    assert np.array_equal(state2["node"]["function_node_result"], [3, 4])


def test_given_value_sampler():
    """Test that we can retrieve numbers from a GivenValueSampler."""
    given_node = GivenValueSampler([1, 3, 5, 7])

    # Check that we have sampled uniformly from the given options.
    state = GraphState(num_samples=5_000)
    results = given_node.compute(state)
    assert len(results) == 5_000
    assert np.all((results == 1) | (results == 3) | (results == 5) | (results == 7))
    assert len(results[results == 1]) > 1000
    assert len(results[results == 3]) > 1000
    assert len(results[results == 5]) > 1000
    assert len(results[results == 7]) > 1000


def test_given_value_sampler_int():
    """Test that we can retrieve numbers from a GivenValueSampler representing a range."""
    given_node = GivenValueSampler(5)

    # Check that we have sampled uniformly from the given options.
    state = GraphState(num_samples=5_000)
    results = given_node.compute(state)
    assert len(results) == 5_000
    assert np.all((results >= 0) & (results < 5))
    for i in range(5):
        assert len(results[results == i]) > 500


def test_given_value_selector():
    """Test that we can retrieve numbers from a GivenValueSelector."""
    index_node = GivenValueList([0, 1, 2, 3, 2, 3, 1, 2], node_label="index_node")
    given_node = GivenValueSelector([10, 20, 30, 40], index_node, node_label="given_node")

    # Check that we have saampled from the given options based on index.
    state = given_node.sample_parameters(num_samples=8)
    assert len(state["given_node"]["function_node_result"]) == 8
    assert np.array_equal(
        state["given_node"]["function_node_result"],
        [10, 20, 30, 40, 30, 40, 20, 30],
    )


def test_given_value_sampler_weighted():
    """Test that we can retrieve numbers from a GivenValueSampler
    with a weighted distribution."""
    given_node = GivenValueSampler([1, 3, 5, 7], [0.1, 0.5, 0.3, 0.1])

    # Check that we have sampled uniformly from the given options
    # with approximately the given weights.
    state = GraphState(num_samples=10_000)
    results = given_node.compute(state)
    assert len(results) == 10_000
    assert np.all((results == 1) | (results == 3) | (results == 5) | (results == 7))
    assert len(results[results == 1]) > 500
    assert len(results[results == 3]) > 4000
    assert len(results[results == 5]) > 2000
    assert len(results[results == 7]) > 500


@pytest.mark.parametrize("test_data_type", ["dict", "ap_table", "pd_df"])
def test_table_sampler(test_data_type):
    """Test that we can retrieve numbers from a TableSampler from a
    dictionary, AstroPy Table, and Panda's DataFrame."""
    raw_data_dict = {
        "A": [1, 2, 3, 4, 5, 6, 7, 8],
        "B": [1, 1, 1, 1, 1, 1, 1, 1],
        "C": [3, 4, 5, 6, 7, 8, 9, 10],
    }

    # Convert the data type depending on the parameterized value.
    if test_data_type == "dict":
        data = raw_data_dict
    elif test_data_type == "ap_table":
        data = Table(raw_data_dict)
    elif test_data_type == "pd_df":
        data = pd.DataFrame(raw_data_dict)
    else:
        data = None

    # Create the table sampler from the data.
    table_node = TableSampler(data, in_order=True, node_label="node")
    state = table_node.sample_parameters(num_samples=2)
    assert len(state) == 3
    assert np.allclose(state["node"]["A"], [1, 2])
    assert np.allclose(state["node"]["B"], [1, 1])
    assert np.allclose(state["node"]["C"], [3, 4])

    state = table_node.sample_parameters(num_samples=1)
    assert len(state) == 3
    assert state["node"]["A"] == 3
    assert state["node"]["B"] == 1
    assert state["node"]["C"] == 5

    state = table_node.sample_parameters(num_samples=4)
    assert len(state) == 3
    assert np.allclose(state["node"]["A"], [4, 5, 6, 7])
    assert np.allclose(state["node"]["B"], [1, 1, 1, 1])
    assert np.allclose(state["node"]["C"], [6, 7, 8, 9])

    # We go past the end of the data.
    with pytest.raises(IndexError):
        _ = table_node.sample_parameters(num_samples=4)

    # We can reset and sample from the beginning.
    table_node.reset()
    state = table_node.sample_parameters(num_samples=2)
    assert len(state) == 3
    assert np.allclose(state["node"]["A"], [1, 2])
    assert np.allclose(state["node"]["B"], [1, 1])
    assert np.allclose(state["node"]["C"], [3, 4])


def test_table_sampler_ranndomized():
    """Test that we can retrieve numbers from a TableSampler."""
    raw_data_dict = {
        "A": [1, 3, 5],
        "B": [2, 4, 6],
    }

    # Create the table sampler from the data.
    table_node = TableSampler(raw_data_dict, node_label="node")
    state = table_node.sample_parameters(num_samples=2000)
    assert len(state) == 2

    # We have sampled the a_vals roughly uniformly from the three options.
    a_vals = state["node"]["A"]
    assert len(a_vals) == 2000
    assert np.all((a_vals == 1) | (a_vals == 3) | (a_vals == 5))
    assert len(a_vals[a_vals == 1]) > 500
    assert len(a_vals[a_vals == 3]) > 500
    assert len(a_vals[a_vals == 5]) > 500

    # We have sampled the b_vals roughly uniformly from the three options.
    b_vals = state["node"]["B"]
    assert len(b_vals) == 2000
    assert np.all((b_vals == 2) | (b_vals == 4) | (b_vals == 6))
    assert len(b_vals[b_vals == 2]) > 500
    assert len(b_vals[b_vals == 4]) > 500
    assert len(b_vals[b_vals == 6]) > 500

    # We always sample consistent ROWS of a and b.
    assert np.all(b_vals - a_vals == 1)



import numpy as np
from astropy.coordinates import angular_separation

from tdastro.sources.physical_model import PhysicalModel


class GaussianGalaxy(PhysicalModel):
    """A galaxy with constant brightness that falls of as a Guassian
    as the distance from the center increases.

    Parameterized values include:
      * brightness - The inherent brightness at the center of the galaxy.
      * dec - The object's declination in degrees. [from PhysicalModel]
      * distance - The object's luminosity distance in pc. [from PhysicalModel]
      * galaxy_radius_std - The standard deviation of the brightness in degrees.
      * ra - The object's right ascension in degrees. [from PhysicalModel]
      * redshift - The object's redshift. [from PhysicalModel]
      * t0 - No effect for a GuassianGalaxy [from PhysicalModel]

    Parameters
    ----------
    brightness : float
        The inherent brightness at the center of the galaxy.
    radius : float
        The standard deviation of the brightness as we move away
        from the galaxy's center (in degrees).
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, brightness, radius, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter("galaxy_radius_std", radius, **kwargs)
        self.add_parameter("brightness", brightness, **kwargs)

    def compute_flux(self, times, wavelengths, graph_state, ra=None, dec=None, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        ra : float, optional
            The right ascension of the observations in degrees.
        dec : float, optional
            The declination of the observations in degrees.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)

        dist = 0.0
        if ra is not None and dec is not None:
            dist = angular_separation(
                params["ra"] * np.pi / 180.0,
                params["dec"] * np.pi / 180.0,
                ra * np.pi / 180.0,
                dec * np.pi / 180.0,
            )

        # Scale the brightness as a Guassian function centered on the object's RA and Dec.
        std = params["galaxy_radius_std"]
        scale = np.exp(-(dist * dist) / (2.0 * std * std))

        return np.full((len(times), len(wavelengths)), params["brightness"] * scale)
